##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.cache.interfaces import IStateCache
from relstorage.tests import TestCase

from . import MockOptions
from . import MockOptionsWithFakeMemcache

class AbstractStateCacheTests(TestCase):
    """
    Tests that all implementations of `IStateCache`
    should be able to pass.
    """

    def getClass(self):
        raise unittest.SkipTest("No implementation defined.")

    Options = MockOptions

    def _makeOne(self, **kw):
        options = self.Options.from_args(**kw)
        inst = self.getClass()(options)
        return inst

    def test_provides(self):
        assert_that(self._makeOne(), validly_provides(IStateCache))

    def test_delitem_not_there(self):
        c = self._makeOne()
        del c[(1, 1)]

    def test_set_all_for_tid(self):
        c = self._makeOne()

        c.set_all_for_tid(
            1,
            [(b'abc', 0, -1),
             (b'ghi', 1, -1),])
        c.set_all_for_tid(
            2,
            [(b'def', 0, -1)]
        )
        # Hits on primary key
        self.assertEqual(c[0, 1],
                         (b'abc', 1))
        self.assertEqual(c[1, 1],
                         (b'ghi', 1))

        self.assertEqual(c[0, 2],
                         (b'def', 2))

    def test_updating_delta_map(self):
        self.assertIs(self._makeOne().updating_delta_map(self), self)

class MemcacheClientTests(AbstractStateCacheTests):

    def setUp(self):
        from relstorage.tests.fakecache import data
        data.clear()

    tearDown = setUp

    Options = MockOptionsWithFakeMemcache

    def getClass(self):
        from relstorage.cache.storage_cache import MemcacheStateCache
        return MemcacheStateCache.from_options

    def test_new_instance_is_really_new(self):
        inst = self._makeOne()
        new = inst.new_instance()
        self.assertIsNot(inst, new)



# -*- coding: utf-8 -*-
"""
Tests for the Oracle dialect.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from relstorage.tests import TestCase
from relstorage.tests import MockCursor

from ...schema import Schema
from ...sql import it

from ..dialect import OracleDialect

class Context(object):
    keep_history = True
    dialect = OracleDialect()

class Driver(object):
    dialect = OracleDialect()
    binary_column_as_state_type = binary_column_as_bytes = lambda b: b
    Binary = bytes

class TestOracleDialect(TestCase):

    def test_boolean_parameter(self):
        transaction = Schema.transaction
        object_state = Schema.object_state

        stmt = transaction.natural_join(
            object_state
        ).select(
            it.c.tid, it.c.username, it.c.description, it.c.extension,
            object_state.c.state_size
        ).where(
            it.c.zoid == it.bindparam("oid")
        ).and_(
            it.c.packed == False # pylint:disable=singleton-comparison
        ).order_by(
            it.c.tid, "DESC"
        )

        # By default we get the generic syntax
        self.assertEqual(
            str(stmt),
            'SELECT tid, username, description, extension, state_size '
            'FROM transaction '
            'JOIN object_state '
            'USING (tid) '
            "WHERE ((zoid = %(oid)s AND packed = FALSE)) "
            'ORDER BY tid DESC'
        )

        # But once we bind to the dialect we get the expected value
        stmt = stmt.bind(Context)
        self.assertEqual(
            str(stmt),
            'SELECT tid, username, description, extension, state_size '
            'FROM transaction '
            'JOIN object_state '
            'USING (tid) '
            "WHERE ((zoid = :oid AND packed = 'N')) "
            'ORDER BY tid DESC'
        )

    def test_boolean_result(self):
        """
        SELECT tid, username, description, extension,

            FROM transaction
            WHERE tid >= 0
        """

        transaction = Schema.transaction
        stmt = transaction.select(
            transaction.c.tid,
            transaction.c.username,
            transaction.c.description,
            transaction.c.extension,
            transaction.c.packed
        ).where(
            transaction.c.tid >= 0
        )

        stmt = stmt.bind(Context)

        self.assertEqual(
            str(stmt),
            'SELECT tid, username, description, extension, '
            "CASE WHEN packed = 'Y' THEN 1 ELSE 0 END "
            'FROM transaction WHERE (tid >= 0)'
        )

class TestMoverQueries(TestCase):

    def _makeOne(self):
        from ..mover import OracleObjectMover
        from relstorage.tests import MockOptions

        return OracleObjectMover(Driver(), MockOptions())

    def test_move_named_query(self):

        inst = self._makeOne()

        unbound = type(inst)._move_from_temp_hf_upsert_query

        self.assertRegex(
            str(unbound),
            r'EXECUTE rs_prep_stmt_.*'
        )
        self.maxDiff = None
        # Bound, the py-format style escapes are replaced with
        # :name params; these are simple numbers.
        self.assertEqual(
            str(inst._move_from_temp_hf_upsert_query),
            'MERGE INTO object_state USING ( '
            'SELECT zoid, :1 tid, COALESCE(LENGTH(state), 0) state_size, state '
            'FROM temp_store ORDER BY zoid) D '
            'ON (object_state.zoid = D.zoid) '
            'WHEN MATCHED THEN UPDATE SET state = D.state, tid = D.tid, state_size = D.state_size '
            'WHEN NOT MATCHED THEN INSERT (zoid, tid, state_size, state) '
            'VALUES (D.zoid, D.tid, D.state_size, D.state)'
        )

class TestDatabaseIteratorQueries(TestCase):

    def _makeOne(self):
        from relstorage.adapters.dbiter import HistoryPreservingDatabaseIterator
        return HistoryPreservingDatabaseIterator(Driver())

    def test_query(self):
        inst = self._makeOne()

        unbound = type(inst)._iter_objects_query

        self.assertEqual(
            str(unbound),
            'SELECT zoid, state FROM object_state WHERE (tid = %(tid)s) ORDER BY zoid'
        )

        # Bound we get pyformat %(name)s replaced with :name
        self.assertEqual(
            str(inst._iter_objects_query),
            'SELECT zoid, state FROM object_state WHERE (tid = :tid) ORDER BY zoid'
        )

        cursor = MockCursor()
        list(inst.iter_objects(cursor, 1)) # Iterator, must flatten

        self.assertEqual(1, len(cursor.executed))
        stmt, params = cursor.executed[0]

        self.assertEqual(
            stmt,
            'SELECT zoid, state FROM object_state WHERE (tid = :tid) ORDER BY zoid'
        )

        self.assertEqual(
            params,
            {'tid': 1}
        )

    def test_iter_transactions(self):
        inst = self._makeOne()
        cursor = MockCursor()

        inst.iter_transactions(cursor)

        self.assertEqual(1, len(cursor.executed))
        stmt, params = cursor.executed[0]

        self.assertEqual(
            stmt,
            'SELECT tid, username, description, extension, 0 '
            'FROM transaction '
            "WHERE ((packed = 'N' AND tid <> 0)) "
            'ORDER BY tid DESC'
        )

        self.assertEqual(
            params,
            None
        )



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Database schema installers
"""
from __future__ import absolute_import
from __future__ import print_function

from zope.interface import implementer

from ..connmanager import connection_callback
from ..interfaces import ISchemaInstaller
from ..schema import AbstractSchemaInstaller

logger = __import__('logging').getLogger(__name__)

class _StoredFunction(object):
    """
    Represents a function stored in the database.

    It is equal to another instance if it has the same
    name and same checksum (signature and creation statement
    are ignored).
    """

    # The function's name
    name = None

    # The function's signature: (boolean, bytea, bytea)
    signature = None

    # The function's CREATE statement
    create = None

    # The checksum for the create statement,
    # as stored in the object's comment.
    checksum = None

    def __init__(self, name, signature, checksum, create=None):
        self.name = name
        self.signature = signature
        self.checksum = checksum
        self.create = create

    def __str__(self):
        """
        Return ``name(signature)``
        """
        if self.signature is None:
            # We didn't read this from the database at all.
            # Don't pretend we know what it is.
            return self.name
        # Signature can be the empty string if it takes no params.
        return "%s(%s)" % (self.name, self.signature)

    def __repr__(self):
        return "<%s %s>" % (
            self, self.checksum
        )

    def __eq__(self, other):
        if not isinstance(other, _StoredFunction):
            return NotImplemented
        return self.name == other.name and self.checksum == other.checksum


@implementer(ISchemaInstaller)
class PostgreSQLSchemaInstaller(AbstractSchemaInstaller):

    database_type = 'postgresql'

    _PROCEDURES = {} # Caching of proc files.

    def __init__(self, options, connmanager, runner, locker):
        self.options = options
        super(PostgreSQLSchemaInstaller, self).__init__(
            connmanager, runner, options.keep_history)
        self.locker = locker

    def _read_proc_files(self):
        procs = super(PostgreSQLSchemaInstaller, self)._read_proc_files()
        # Convert from bare strings into _StoredFunction objects
        # (which are missing their signatures at this point).
        current_object = 'current_object' if self.keep_history else 'object_state'
        if self.keep_history:
            object_state_join = """
            INNER JOIN object_state ON (object_state.zoid = cur.zoid
                                       AND object_state.tid = cur.tid)
            """
            object_state_name = 'object_state'
        else:
            object_state_join = ""
            object_state_name = 'cur'
        return {
            name: _StoredFunction(
                name,
                None,
                self._checksum_for_str(value),
                value.format(
                    CURRENT_OBJECT=current_object,
                    OBJECT_STATE_JOIN=object_state_join,
                    OBJECT_STATE_NAME=object_state_name
                )
            )
            for name, value
            in procs.items()
        }

    def get_database_name(self, cursor):
        cursor.execute("SELECT current_database()")
        row, = cursor.fetchall()
        name, = row
        return self._metadata_to_native_str(name)

    @connection_callback(inherit=AbstractSchemaInstaller._prepare_with_connection)
    def _prepare_with_connection(self, conn, cursor):
        super(PostgreSQLSchemaInstaller, self)._prepare_with_connection(conn, cursor)

        # Do we need to merge blob chunks?
        if not self.options.shared_blob_dir:
            cursor.execute('SELECT chunk_num FROM blob_chunk WHERE chunk_num > 0 LIMIT 1')
            if cursor.fetchone():
                logger.info("Merging blob chunks on the server.")
                cursor.execute("SELECT merge_blob_chunks()")
                # If we've done our job right, any blobs cached on
                # disk are still perfectly valid.

    def create_procedures(self, cursor):
        if not self.__all_procedures_installed(cursor):
            self.__install_procedures(cursor)
            if not self.__all_procedures_installed(cursor):
                raise AssertionError(
                    "Could not get version information after "
                    "installing the stored procedures.")

    def create_triggers(self, cursor):
        triggers = self.list_triggers(cursor)
        __traceback_info__ = triggers, self.list_tables(cursor), self.get_database_name(cursor)
        if 'blob_chunk_delete' not in triggers:
            self.__install_triggers(cursor)

    def __native_names_only(self, cursor):
        native = self._metadata_to_native_str
        return frozenset([
            native(name)
            for (name,) in cursor.fetchall()
        ])

    def list_tables(self, cursor):
        cursor.execute("SELECT tablename FROM pg_tables")
        return self.__native_names_only(cursor)

    def list_sequences(self, cursor):
        cursor.execute("SELECT relname FROM pg_class WHERE relkind = 'S'")
        return self.__native_names_only(cursor)

    def list_views(self, cursor):
        cursor.execute("SELECT relname FROM pg_class WHERE relkind = 'v'")
        return self.__native_names_only(cursor)

    def list_languages(self, cursor):
        cursor.execute("SELECT lanname FROM pg_catalog.pg_language")
        return self.__native_names_only(cursor)

    def __install_languages(self, cursor):
        if 'plpgsql' not in self.list_languages(cursor):
            cursor.execute("CREATE LANGUAGE plpgsql")

    def list_procedures(self, cursor):
        """
        Returns {procedure name: _StoredFunction}.
        """
        # The description is populated with
        # ``COMMENT ON FUNCTION <name>(<args>) IS 'comment'``.
        # Prior to Postgres 10, the args are required; with
        # 10 and later they are only needed if the function is overloaded.
        stmt = """
        SELECT p.proname AS funcname,
               d.description,
               pg_catalog.pg_get_function_identity_arguments(p.oid)
        FROM pg_proc p
        INNER JOIN pg_namespace n ON n.oid = p.pronamespace
        LEFT JOIN pg_description As d ON (d.objoid = p.oid)
        WHERE n.nspname = 'public'
        """

        cursor.execute(stmt)
        res = {}
        native = self._metadata_to_native_str
        for (name, checksum, signature) in cursor.fetchall():
            name = native(name)
            checksum = native(checksum)
            signature = native(signature)

            disk_func = self.procedures.get(name)
            res[name.lower()] = db_func = _StoredFunction(
                name,
                signature,
                checksum,
                # Include the source, if it's around
                disk_func.create if disk_func else None
            )

            if disk_func and disk_func.checksum == db_func.checksum:
                # Yay, what's in the database matches what's read from disk.
                # That means the signature must match too.
                disk_func.signature = db_func.signature
                # But of course we don't add a checksum to the db_func
                # if we don't already have one; we wouldn't be able to tell
                # a mismatch.

        return res

    def __all_procedures_installed(self, cursor):
        """
        Check whether all required stored procedures are installed.

        Returns True only if all required procedures are installed and
        up to date.
        """

        expected = self.procedures

        installed = self.list_procedures(cursor)
        # If the database evolves over tiem, there could be
        # extra procs still there that we don't care about.
        installed = {k: v for k, v in installed.items() if k in expected}
        if installed != expected:
            logger.info(
                "Procedures incorrect, will reinstall. "
                "Expected: %s."
                "Actual: %s",
                expected, installed
            )
            return False
        return True

    def __install_procedures(self, cursor):
        """Install the stored procedures"""
        self.__install_languages(cursor)

        # PostgreSQL procedures in the SQL language
        # do lots of validation at compile time; in particular,
        # they check that the functions they use in SELECT statements
        # actually exist. When we have procedures that call each other,
        # that means there's an order they have to be created in.
        # Rather than try to figure out what that order is, or encode it
        # in names somehow, we'll disable that validation for the duration
        # of this transaction (just like pg_dump.)
        cursor.execute('SET LOCAL check_function_bodies = off')

        for proc_name, stored_func in self.procedures.items():
            __traceback_info__ = proc_name, self.keep_history
            proc_source = stored_func.create
            # All definitions should be written with 'CREATE OR REPLACE'
            # so we don't need to bother with 'DROP'. Though, if the return
            # type changes, we can't REPLACE.
            cursor.execute(proc_source)

        # Update checksums
        # Postgres < 10 requires the signature to identify the function;
        # after that it's optional if the function isn't overloaded.
        # Rather than try to parse the signature from the file ourself, we
        # let the database do it and then ask it.
        for db_proc in self.list_procedures(cursor).values():
            # db_proc will have the signature but perhaps not the checksum.
            try:
                disk_proc = self.procedures[db_proc.name]
            except KeyError: # pragma: no cover
                # One in the DB no longer on disk.
                # Ignore.
                continue
            disk_proc.signature = db_proc.signature
            if db_proc.checksum == disk_proc.checksum:
                continue

            db_proc.checksum = disk_proc.checksum

            # For pg8000 we can't use a parameter here (because it prepares?)
            comment = "COMMENT ON FUNCTION %s IS '%s'" % (
                str(db_proc), db_proc.checksum
            )
            __traceback_info__ = comment
            cursor.execute(comment)


    def list_triggers(self, cursor):
        cursor.execute("SELECT tgname FROM pg_trigger")
        return self.__native_names_only(cursor)

    def __install_triggers(self, cursor):
        stmt = """
        CREATE TRIGGER blob_chunk_delete
            BEFORE DELETE ON blob_chunk
            FOR EACH ROW
            EXECUTE PROCEDURE blob_chunk_delete_trigger()
        """
        cursor.execute(stmt)

    def drop_all(self):
        def delete_blob_chunk(_conn, cursor):
            if 'blob_chunk' in self.list_tables(cursor):
                # Trigger deletion of blob OIDs.
                cursor.execute("DELETE FROM blob_chunk")
        self.connmanager.open_and_call(delete_blob_chunk)
        super(PostgreSQLSchemaInstaller, self).drop_all()

    def _create_pack_lock(self, cursor):
        return

    def _create_new_oid(self, cursor):
        stmt = """
        CREATE SEQUENCE IF NOT EXISTS zoid_seq;
        """
        self.runner.run_script(cursor, stmt)


    CREATE_PACK_OBJECT_IX_TMPL = """
    CREATE INDEX pack_object_keep_false ON pack_object (zoid)
        WHERE keep = false;
    CREATE INDEX pack_object_keep_true ON pack_object (visited)
        WHERE keep = true;
    """

    DROP_TABLE_TMPL = 'DROP TABLE IF EXISTS {table}'

    def _reset_oid(self, cursor):
        stmt = "ALTER SEQUENCE zoid_seq RESTART WITH 1;"
        self.runner.run_script(cursor, stmt)

    # Use the fast, semi-transactional way to truncate tables. It's
    # not MVCC safe, but "TRUNCATE is transaction-safe with respect to
    # the data in the tables: the truncation will be safely rolled
    # back if the surrounding transaction does not commit."
    _zap_all_tbl_stmt = 'TRUNCATE TABLE %s CASCADE'

    def _before_zap_all_tables(self, cursor, tables, slow=False):
        super(PostgreSQLSchemaInstaller, self)._before_zap_all_tables(cursor, tables, slow)
        if not slow and 'blob_chunk' in tables:
            # If we're going to be truncating, it's important to
            # remove the large objects through lo_unlink. We have a
            # trigger that does that, but only for DELETE.
            # The `vacuumlo` command cleans up any that might have been
            # missed.

            # This unfortunately results in returning a row for each
            # object unlinked, but it should still be faster than
            # running a DELETE and firing the trigger for each row.
            cursor.execute("""
            SELECT lo_unlink(t.chunk)
            FROM
            (SELECT DISTINCT chunk FROM blob_chunk)
            AS t
            """)



#!/usr/bin/env python
##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
ZODB storage packing utility.
"""

import argparse
import logging
import sys
import time
from io import StringIO

import ZConfig
import ZODB.serialize

schema_xml = u"""
<schema>
  <import package="ZODB"/>
  <import package="relstorage"/>
  <multisection type="ZODB.storage" attribute="storages" />
</schema>
"""

logger = logging.getLogger("zodbpack")


def main(argv=None):
    if argv is None:
        argv = sys.argv
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        "-d", "--days", dest="days", default=1,
        help="Days of history to keep (default: %(default)s)",
        type=float,
    )
    parser.add_argument(
        "--prepack", dest="prepack", default=False,
        action="store_true",
        help="Perform only the pre-pack preparation stage of a pack. "
        "(Only works with some storage types)",
    )
    parser.add_argument(
        "--use-prepack-state", dest="reuse_prepack", default=False,
        action="store_true",
        help="Skip the preparation stage and go straight to packing. "
        "Requires that a pre-pack has been run, or that packing was aborted "
        "before it was completed.",
    )
    parser.add_argument("config_file", type=argparse.FileType('r'))
    options = parser.parse_args(argv[1:])

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s [%(name)s] %(levelname)s %(message)s"
    )

    schema = ZConfig.loadSchemaFile(StringIO(schema_xml))
    config, _ = ZConfig.loadConfigFile(schema, options.config_file)

    t = time.time() - options.days * 86400.0
    for s in config.storages:
        name = '%s (%s)' % ((s.name or 'storage'), s.__class__.__name__)
        logger.info("Opening %s...", name)
        storage = s.open()
        logger.info("Packing %s.", name)
        if options.prepack or options.reuse_prepack:
            storage.pack(t, ZODB.serialize.referencesf,
                         prepack_only=options.prepack,
                         skip_prepack=options.reuse_prepack)
        else:
            # Be non-relstorage Storages friendly
            storage.pack(t, ZODB.serialize.referencesf)
        storage.close()
        logger.info("Packed %s.", name)

if __name__ == '__main__':
    main()




import unittest


class TestTreeMarker(unittest.TestCase):

    @property
    def _class(self):
        from ..treemark import TreeMarker
        return TreeMarker

    def _make(self):
        return self._class()

    def test_simple_tree(self):
        obj = self._make()
        obj.add_refs([
            (5, 7),
            (5, 9),
            (6, 1),
        ])
        obj.mark([5])
        self.assertEqual(set(obj.reachable), set([5, 7, 9]))

    def test_multiple_passes_required(self):
        obj = self._make()
        obj.add_refs([
            (3, 1),
            (5, 7),
            (5, 9),
            (6, 1),
            (9, 3),
            (9, 13),
        ])
        obj.mark([5])
        self.assertEqual(set(obj.reachable), set([1, 3, 5, 7, 9, 13]))

    def test_circular(self):
        obj = self._make()
        obj.add_refs([
            (5, 7),
            (5, 8),
            (7, 5),
            (8, 9),
            (9, 5),
            (10, 1),
        ])
        obj.mark([5])
        self.assertEqual(set(obj.reachable), set([5, 7, 8, 9]))

    def test_64bit(self):
        obj = self._make()
        obj.add_refs([
            (5, 7),
            (5, 9 << 32),
            (6 << 32, 1),
        ])
        obj.mark([5])
        self.assertEqual(set(obj.reachable), set([5, 7, 9 << 32]))

    def test_64bit_with_multiple_passes_required(self):
        obj = self._make()
        obj.add_refs([
            (3 << 32, 1),
            (5, 7),
            (5, 9 << 32),
            (6 << 32, 1),
            (9 << 32, 3 << 32),
            (9 << 32, 13),
        ])
        obj.mark([5])
        self.assertEqual(set(obj.reachable), set([
            1, 3 << 32, 5, 7, 9 << 32, 13]))

    def test_64bit_circular(self):
        obj = self._make()
        obj.add_refs([
            (5, 7),
            (5, 8 << 32),
            (7, 5),
            (8 << 32, 9 << 32),
            (9 << 32, 5),
            (10 << 32, 1),
        ])
        obj.mark([5])
        self.assertEqual(set(obj.reachable), set([5, 7, 8 << 32, 9 << 32]))

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestTreeMarker))
    return suite

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



"""This package contains test code copied from ZODB 3.9 with minor alterations.

It is especially useful for testing RelStorage + ZODB 3.8.
"""
import os
from nti.testing.time import MonotonicallyIncreasingTimeLayerMixin

from ZODB.DB import DB
from ZODB.tests.util import setUp
from ZODB.tests.util import tearDown

class TestBlobMixin(object):

    DEFAULT_BLOB_STORAGE_KWARGS = {}

    def setUp(self):
        super(TestBlobMixin, self).setUp()
        setUp(self)
        self._timer = MonotonicallyIncreasingTimeLayerMixin()
        self._timer.testSetUp()
        try:
            self.blob_storage = self.create_storage(**self.DEFAULT_BLOB_STORAGE_KWARGS)
        except: # pragma: no cover
            # If setUp() raises an exception, tearDown is never called.
            # That's bad: ZODB.tests.util.setUp() changes directories and
            # monkeys with the contents of the stdlib tempfile.
            tearDown(self)
            raise
        self.database = DB(self.blob_storage)

    def tearDown(self):
        self.database.close()
        self.blob_storage.close()
        self._timer.testTearDown()
        tearDown(self)
        super(TestBlobMixin, self).tearDown()

    def _count_and_size_blobs_in_directory(self):
        size = 0
        count = 0
        d = self.blob_storage.blobhelper.blob_dir
        for base, _, files in os.walk(d):
            for f in files:
                if f.endswith('.blob'):
                    count += 1
                    try:
                        size += os.stat(os.path.join(base, f)).st_size
                    except OSError: # pragma: no cover
                        if os.path.exists(os.path.join(base, f)):
                            raise
        return count, size

    def _size_blobs_in_directory(self):
        return self._count_and_size_blobs_in_directory()[1]

    def _count_blobs_in_directory(self):
        return self._count_and_size_blobs_in_directory()[0]




"""Test a big tree of random OID references.

Run like this: bin/py -m cProfile -- -s tottime relstorage/tests/bigmark.py
"""

import logging
from random import randint
from random import random

from relstorage._compat import xrange
from relstorage.treemark import TreeMarker

log = logging.getLogger('bigmark')


def bigmark():
    """Follow ~20M references between 10M objects."""

    logging.basicConfig(
        level=logging.DEBUG,
        format='%(asctime)s %(levelname)s [%(name)s] %(message)s')

    oid_count = 10 * 1000 * 1000
    k = 10000

    log.info("Generating random references.")

    marker = TreeMarker()
    marker.add_refs([(0, i * k) for i in range(20)])
    refcount = 20
    for i in xrange(1, oid_count):
        if random() < 0.2:
            refs = []
            for _ in range(randint(0, 20)):
                refs.append((i * k, randint(0, oid_count) * k))
            marker.add_refs(refs)
            refcount += len(refs)

    log.info("Generated %d references.", refcount)

    log.info("Finding reachable objects.")

    pass_count = marker.mark([0])

    log.info(
        "Found %d reachable objects in %d passes.",
        marker.reachable_count, pass_count)


if __name__ == '__main__':
    bigmark()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import implementer

from ..interfaces import ILocker
from ..locker import AbstractLocker


@implementer(ILocker)
class Sqlite3Locker(AbstractLocker):

    supports_row_lock_nowait = False
    _lock_share_clause = '<This should not be used>'
    _lock_share_clause_nowait = '<This should not be used>'
    _lock_current_clause = '<This Should not be used>'

    def _lock_readCurrent_oids_for_share(self, cursor, current_oids, shared_locks_block):
        """
        This is a no-op.

        Once this is called, the parent class will call the after_lock_share callback,
        and the parent adapter will use that to check the recorded TIDs against the
        desired TIDs. Our store connection is still in autocommit mode at this point,
        so it will read the current data.
        """

    _lock_current_objects_query = 'UPDATE commit_row_lock SET tid = tid WHERE tid < 0'

    def _lock_rows_being_modified(self, cursor):
        # We exclusively lock the whole database at this point. There's no way to
        # avoid it. We also switch out of autocommit mode and start a transaction,
        # meaning our store cursor is no longer updating...but since we have an exclusive
        # lock, that's fine. We use a statement that matches no rows, but still takes
        # locks.
        conn = cursor.connection
        # If we pre-allocated the TID due to a restore, we could already be
        # in a tranasction. In which case there's nothing to do (but for safety
        # we still execute the query)
        AbstractLocker._lock_rows_being_modified(self, cursor)
        assert conn.in_transaction or conn.in_transaction is None, repr(conn) # Py2

    def hold_commit_lock(self, cursor, ensure_current=False, nowait=False):
        # We may or may not actually have locked rows; if we're doing a restore
        # we allocate the tid first. Likewise, an undo also
        # locks upfront.
        if ensure_current:
            cursor.execute(self._lock_current_objects_query)
            in_transaction = cursor.connection.in_transaction
            assert in_transaction or in_transaction is None # Py2

    def release_commit_lock(self, cursor):
        # no action needed, locks released with transaction.
        pass

    def release_pack_lock(self, cursor):
        pass

    def hold_pack_lock(self, cursor):
        pass



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import print_function

from collections import namedtuple

from zope.interface import implementer

from ZODB.utils import p64 as int64_to_8bytes

from relstorage._compat import TidList
from relstorage._util import Lazy

from ._util import DatabaseHelpersMixin
from .interfaces import IDatabaseIterator
from .schema import Schema
from .sql import it

class DatabaseIterator(DatabaseHelpersMixin):
    """
    Abstract base class for database iteration.
    """

    def __init__(self, database_driver):
        """
        :param database_driver: Necessary to bind queries correctly.
        """
        self.driver = database_driver

    @Lazy
    def _as_state(self):
        return self.driver.binary_column_as_state_type

    @Lazy
    def _as_bytes(self):
        return self.driver.binary_column_as_bytes

    _iter_objects_query = Schema.object_state.select(
        it.c.zoid,
        it.c.state
    ).where(
        it.c.tid == it.bindparam('tid')
    ).order_by(
        it.c.zoid
    )

    def iter_objects(self, cursor, tid):
        """Iterate over object states in a transaction.

        Yields ``(oid, state)`` for each object in the transaction.
        """
        self._iter_objects_query.execute(cursor, {'tid': tid})
        as_state = self._as_state
        for oid, state in cursor:
            state = as_state(state) # pylint:disable=too-many-function-args
            yield oid, state

    _iter_current_records = Schema.all_current_object_state.select(
        it.c.zoid,
        it.c.tid,
        it.c.state
    ).where(
        it.c.zoid >= it.bindparam('start_oid')
    ).order_by(
        it.c.zoid
    )

    def iter_current_records(self, cursor, start_oid_int=0):
        """
        Cause the *cursor* (which should be a server-side cursor)
        to execute a query that will iterate over
        ``(oid_int, tid_int, state_bytes)`` values for all the current objects.

        Each current object is returned only once, at the transaction most recently
        committed for it.

        Returns a generator.
        """
        self._iter_current_records.execute(cursor, {'start_oid': start_oid_int})
        i_b = int64_to_8bytes
        s = self._as_state
        for oid_int, tid_int, state_bytes in cursor:
            yield i_b(oid_int), i_b(tid_int), s(state_bytes) # pylint:disable=too-many-function-args


class _HistoryPreservingTransactionRecord(namedtuple(
        '_HistoryPreservingTransactionRecord',
        ('tid_int', 'username', 'description', 'extension', 'packed')
)):
    __slots__ = ()

    @property
    def pickle_size(self):
        return self.packed


@implementer(IDatabaseIterator)
class HistoryPreservingDatabaseIterator(DatabaseIterator):

    keep_history = True

    def _transaction_iterator(self, cursor):
        """
        Iterate over a list of transactions returned from the database.

        Each row is ``(tid, username, description, extension, X)``
        """
        # Iterating the cursor itself in a generator is not safe if
        # the cursor doesn't actually buffer all the rows *anyway*. If
        # we break from the iterating loop before exhausting all the
        # rows, a subsequent query or close operation can lead to
        # things like MySQL Connector/Python raising
        # InternalError(unread results)
        # Because we have it all in memory anyway, there's not much point in
        # making this a generator.

        # Although the transaction interface for username and description are
        # defined as strings, this layer works with bytes. The ZODB layer
        # does the conversion.
        as_bytes = self._as_bytes
        # pylint:disable=too-many-function-args
        return [
            _HistoryPreservingTransactionRecord(
                tid,
                as_bytes(username),
                as_bytes(description),
                as_bytes(ext),
                packed
            )
            for (tid, username, description, ext, packed)
            in cursor
        ]

    _iter_transactions_query = Schema.transaction.select(
        it.c.tid, it.c.username, it.c.description, it.c.extension, 0
    ).where(
        it.c.packed == False # pylint:disable=singleton-comparison
    ).and_(
        it.c.tid != 0
    ).order_by(
        it.c.tid, 'DESC'
    )

    def iter_transactions(self, cursor):
        """Iterate over the transaction log, newest first.

        Skips packed transactions.
        Yields (tid, username, description, extension) for each transaction.
        """
        self._iter_transactions_query.execute(cursor)
        return self._transaction_iterator(cursor)

    _iter_transactions_range_query = Schema.transaction.select(
        it.c.tid,
        it.c.username,
        it.c.description,
        it.c.extension,
        it.c.packed,
    ).where(
        it.c.tid >= it.bindparam('min_tid')
    ).and_(
        it.c.tid <= it.bindparam('max_tid')
    ).order_by(
        it.c.tid
    )

    def iter_transactions_range(self, cursor, start=None, stop=None):
        """
        See `IDatabaseIterator`.
        """
        params = {
            'min_tid': start if start else 0,
            'max_tid': stop if stop else self.MAX_TID
        }
        self._iter_transactions_range_query.execute(cursor, params)
        return self._transaction_iterator(cursor)

    _object_exists_query = Schema.current_object.select(
        1
    ).where(
        it.c.zoid == it.bindparam('oid')
    )

    _object_history_query = Schema.transaction.natural_join(
        Schema.object_state
    ).select(
        it.c.tid, it.c.username, it.c.description, it.c.extension,
        Schema.object_state.c.state_size
    ).where(
        it.c.zoid == it.bindparam("oid")
    ).and_(
        it.c.packed == False # pylint:disable=singleton-comparison
    ).order_by(
        it.c.tid, "DESC"
    )

    def iter_object_history(self, cursor, oid):
        """
        See `IDatabaseIterator`
        Raises KeyError if the object does not exist.
        """
        params = {'oid': oid}
        self._object_exists_query.execute(cursor, params)
        if not cursor.fetchall():
            raise KeyError(oid)

        self._object_history_query.execute(cursor, params)
        return self._transaction_iterator(cursor)

class _HistoryFreeTransactionRecord(object):
    __slots__ = ('tid_int',)

    username = b''
    description = b''
    extension = b''
    packed = True

    def __init__(self, tid):
        self.tid_int = tid


class _HistoryFreeObjectHistoryRecord(_HistoryFreeTransactionRecord):
    __slots__ = ('pickle_size',)

    def __init__(self, tid, size):
        _HistoryFreeTransactionRecord.__init__(self, tid)
        self.pickle_size = size


class _HistoryFreeTransactionRange(object):
    # By storing just the int, and materializing the records on demand, we
    # save substantial amounts of memory. For example, 18MM records on PyPy
    # went from about 3.5GB to about 0.5GB
    __slots__ = (
        'tid_ints',
    )

    def __init__(self, tid_ints):
        self.tid_ints = tid_ints

    def __len__(self):
        return len(self.tid_ints)

    def __getitem__(self, ix):
        return _HistoryFreeTransactionRecord(self.tid_ints[ix])

@implementer(IDatabaseIterator)
class HistoryFreeDatabaseIterator(DatabaseIterator):

    keep_history = False

    def iter_transactions(self, cursor):
        """
        This always returns an empty iterable.
        """
        # pylint:disable=unused-argument
        return ()

    _iter_transactions_range_query = Schema.object_state.select(
        it.c.tid,
    ).where(
        it.c.tid >= it.bindparam('min_tid')
    ).and_(
        it.c.tid <= it.bindparam('max_tid')
    ).order_by(
        it.c.tid
    ).distinct()

    def iter_transactions_range(self, cursor, start=None, stop=None):
        """
        See `IDatabaseIterator`.
        """
        params = {
            'min_tid': start if start else 0,
            'max_tid': stop if stop else self.MAX_TID
        }
        self._iter_transactions_range_query.execute(cursor, params)
        return _HistoryFreeTransactionRange(TidList((tid for (tid,) in cursor)))

    _iter_object_history_query = Schema.object_state.select(
        it.c.tid, it.c.state_size
    ).where(
        it.c.zoid == it.bindparam('oid')
    )

    def iter_object_history(self, cursor, oid):
        """
        See `IDatabaseIterator`

        Yields a single row.
        """
        self._iter_object_history_query.execute(cursor, {'oid': oid})
        rows = cursor.fetchall()
        if not rows:
            raise KeyError(oid)
        assert len(rows) == 1
        tid, size = rows[0]
        return [_HistoryFreeObjectHistoryRecord(tid, size)]



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from zope.interface import implementer

from relstorage._util import consume

from ..interfaces import IObjectMover
from ..schema import Schema
from .scriptrunner import Sqlite3ScriptRunner
from ..mover import AbstractObjectMover
from ..mover import metricmethod_sampled

from .batch import Sqlite3RowBatcher


@implementer(IObjectMover)
class Sqlite3ObjectMover(AbstractObjectMover):

    _create_temp_store = Schema.temp_store.create()
    # SQLite doesn't do well at joining temporary tables to normal tables.
    # Even after running ANALYZE. (ANALYZE doesn't get temp tables). It assumes
    # that a table with no stats has a million rows (1,048,576), so until
    # the normal table is actually recorded as having a million rows, the join
    # order of these two tables is always going to put OBJECT_STATE as the outer loop
    # and scan that table. In practice, that's a poor choice: object_state will have
    # many more rows than the temp table. We can workaround that by explicitly
    # adding a stat for object_state to sqlite_stat1, but that will vanish
    # the next time an ANALYZE is run. Instead, if we use a CROSS JOIN,
    # we force the optimizer to follow our query ordering and make the smaller
    # number of loops.
    #
    # References:
    # - https://www.sqlite.org/lang_select.html#fromclause
    # - https://www.sqlite.org/optoverview.html#manual_control_of_query_plans_using_cross_join
    _detect_conflict_query = AbstractObjectMover._detect_conflict_query.join_kind(
        'CROSS JOIN'
    )

    def __init__(self, database_driver, options, runner=None,
                 version_detector=None,
                 batcher_factory=Sqlite3RowBatcher):
        super(Sqlite3ObjectMover, self).__init__(
            database_driver,
            options, runner=runner, version_detector=version_detector,
            batcher_factory=batcher_factory)

        # We only ever have one blob chunk so we can simplify that
        # table.
        self.__on_store_opened_script = """
        CREATE TEMPORARY TABLE IF NOT EXISTS temp_blob_chunk (
                zoid        INTEGER PRIMARY KEY,
                chunk_num   INTEGER NOT NULL,
                chunk       BLOB,
                CHECK(chunk_num = 0)
        );
        """ + str(self._create_temp_store) + """;
        """

    @metricmethod_sampled
    def on_store_opened(self, cursor, restart=False):
        """Create the temporary table for storing objects"""
        # Don't use cursor.executescript, it commits. Remember our store
        # connection is usually meant to be in auto-commit mode and shouldn't take
        # exclusive locks until tpc_vote time. Prior to Python 3.6, if we were
        # in a transaction this would commit it. but we shouldn't be in a transaction.
        assert not cursor.connection.in_transaction
        if not restart:
            Sqlite3ScriptRunner().run_script(cursor, self.__on_store_opened_script)
            assert not cursor.connection.in_transaction
            cursor.connection.register_before_commit_cleanup(self._before_commit)

        super(Sqlite3ObjectMover, self).on_store_opened(cursor, restart)

    def _before_commit(self, connection, _rolling_back=None):
        # Regardless of whether we're rolling back or not we need to delete
        # everything to freshen the connection. DELETE with no WHERE clause
        # in SQLite is optimized like a TRUNCATE.
        consume(connection.execute('DELETE FROM temp_store'))
        consume(connection.execute('DELETE FROM temp_blob_chunk'))

    _upload_blob_uses_chunks = False

    @metricmethod_sampled
    def restore(self, cursor, batcher, oid, tid, data):
        self._generic_restore(batcher, oid, tid, data,
                              command='INSERT OR REPLACE', suffix='')

    def store_temps(self, cursor, state_oid_tid_iter):
        AbstractObjectMover.store_temps(self, cursor, state_oid_tid_iter)
        # That should have started a transaction, effectively moving us
        # from READ COMMITTED mode into REPEATABLE READ. We don't want that because we haven't
        # taken out an exclusive lock on the main database yet (but we do
        # want the write to be transactional and fast) so we now COMMIT and go back to
        # floating.
        cursor.connection.return_to_repeatable_read()

    def replace_temps(self, cursor, state_oid_tid_iter):
        # This is called when we're already holding exclusive locks
        # on the DB and must not release them.
        AbstractObjectMover.store_temps(self, cursor, state_oid_tid_iter)



# -*- coding: utf-8 -*-
"""
Supporting code.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from functools import partial

from zope.interface import implementer

from relstorage._compat import wraps

from .interfaces import IStaleAware

def storage_method(meth):
    """
    Decorator to mark the decorated method as being a storage method,
    to be copied to the storage instance.
    """
    meth.storage_method = True
    return meth

def writable_storage_method(meth):
    """
    Decorator to mark the decorated method as being a storage method,
    to be copied to the storage instance when it is *not* read only.

    If the storage is read only, then calling this method will
    immediately raise a ``ReadOnlyError``.
    """
    meth = storage_method(meth)
    meth.write_method = True
    return meth

def phase_dependent(meth):
    """
    Decorator to mark the method as requiring the current
    transaction phase as its first argument.

    These are methods that can only be called after ``tpc_begin``
    and delegate to the ``tpc_phase`` object.
    """
    meth.phase_dependent = True
    return meth

def phase_dependent_aborts_early(meth):
    """
    Like :func:`phase_dependent`, but if any exception
    were to propagate out of the body of the method,
    the TPC will be aborted  immediately.
    """
    meth = phase_dependent(meth)
    meth.aborts_early = True
    return meth

class _StaleMethodWrapperType(type):
    # Make the _StaleMethodWrapper class also a descriptor
    # so that when we put it in the class dictionary, it
    # gets a chance to bind to the instance.
    def __get__(cls, inst, klass):
        return partial(cls, inst)

@implementer(IStaleAware)
class _StaleMethodWrapper(object):

    __slots__ = (
        'orig_method',
        'stale_error',
    )

    def __init__(self, orig_method, stale_error):
        self.orig_method = orig_method
        self.stale_error = stale_error

    def __call__(self, *args, **kwargs):
        raise self.stale_error

    def no_longer_stale(self):
        return self.orig_method

    def stale(self, stale_error): # pylint:disable=unused-argument
        return self

_StaleMethodWrapper = _StaleMethodWrapperType(
    '_StaleMethodWrapper',
    (object,),
    {
        k: v
        for k, v in _StaleMethodWrapper.__dict__.items()
        if k not in _StaleMethodWrapper.__slots__
    }
)

class _StaleAwareMethodTemplate(object):

    __slots__ = (
        '_wrapped',
    )

    def __init__(self, wrapped):
        self._wrapped = wrapped

    def __getattr__(self, name):
        return getattr(self._wrapped, name)

    def no_longer_stale(self):
        return self

    stale = _StaleMethodWrapper

class stale_aware(object):

    __slots__ = (
        '_func',
    )

    def __init__(self, func):
        self._func = func

    def __get__(self, inst, klass):
        if inst is None:
            return self

        bound = self._func.__get__(inst, klass)
        # Because we cannot add attributes to a bound method,
        # and because we want to limit the overhead of additional
        # function calls (these methods are commonly used),
        # we derive a new class whose __call__ is exactly the bound
        # method's __call__.  For this reason, it's important to access
        # these methods only once and cache them (which is what copy_storage_methods
        # does).

        # Unfortunately, when a method from inside the defining class
        # calls another method decorated like this, we derive a new
        # type each time. e.g., ``Loader.loadBefore`` calls
        # ``self.load`` and generates a new type. Therefore we must
        # also cache this on the instance. This creates a reference
        # cycle.

        # Many of the instances that use this actually define __slots__
        # so we can't put arbitrary stuff in their missing __dict__ anyway,
        # so they must also define __dict__.

        stale_aware_class = type(
            'StaleAware_' + bound.__name__,
            (_StaleAwareMethodTemplate,),
            {
                '__slots__': (),
                '__call__': bound.__call__,
                '__name__': bound.__name__,
                '__module__': bound.__module__,
            }
        )

        stale_aware_class = implementer(IStaleAware)(stale_aware_class)
        # Defining
        meth = stale_aware_class(bound)
        inst.__dict__[bound.__name__] = meth
        return meth


def _make_phase_dependent(storage, method):
    aborts_early = getattr(method, 'aborts_early', False)
    if aborts_early:
        @wraps(method)
        def state(*args, **kwargs):
            try:
                return method(storage._tpc_phase, *args, **kwargs)
            except:
                storage.tpc_abort(None, _force=True)
                raise
    else:
        @wraps(method)
        def state(*args, **kwargs):
            return method(storage._tpc_phase, *args, **kwargs)
    return state

def make_cannot_write(storage, bound_method):
    @wraps(bound_method)
    def read_only(*args, **kwargs):
        raise storage._read_only_error
    return read_only

def copy_storage_methods(storage, delegate):
    # type: (RelStorage, Any) -> None

    read_only = storage.isReadOnly()

    for var_name in dir(delegate):
        __traceabck_info__ = var_name
        if var_name.startswith('_'):
            continue

        attr = getattr(delegate, var_name)
        if callable(attr) and getattr(attr, 'storage_method', None):
            storage_meth = attr

            method_writes_to_db = getattr(storage_meth, 'write_method', False)
            method_is_phase_dependent = getattr(storage_meth, 'phase_dependent', False)

            if read_only and method_writes_to_db:
                storage_meth = make_cannot_write(storage, storage_meth)
            elif method_is_phase_dependent:
                storage_meth = _make_phase_dependent(storage, storage_meth)
            setattr(storage, var_name, storage_meth)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
psycopg2cffi IDBDriver implementations.
"""

from __future__ import absolute_import
from __future__ import print_function

from .psycopg2 import Psycopg2Driver

__all__ = [
    'Psycopg2cffiDriver',
    # TODO: A gevent driver seems like it should be possible?
]

class Psycopg2cffiDriver(Psycopg2Driver):
    __name__ = 'psycopg2cffi'
    MODULE_NAME = __name__
    PRIORITY = 2
    PRIORITY_PYPY = 1

    def _create_connection(self, mod): # pylint:disable=arguments-differ
        # The psycopg2cffi cursor doesn't go through the generic
        # ``execute()`` function for copy_to/from/expert, it goes
        # right to the low-level _pq_execute() method. As a
        # consequence, the connection object's state may not be right.
        # It doesn't know that a transaction has been started, and so
        # consequently the next time we *do* use ``execute()``, it begins a
        # new transaction. Because that clears temp tables, we're hosed.
        # Thus, to make copy play right, we need to keep the connection in the loop;
        # we can do that by making sure we call ``execute()`` before doing a copy,
        # or we can directly make the same low-level call that is missing;
        # we choose the later.

        if getattr(mod, 'RSPsycopg2cffiConnection', self) is self:
            class Cursor(mod.extensions.cursor):

                def copy_expert(self, *args, **kwargs):
                    self.connection._begin_transaction()
                    return super(Cursor, self).copy_expert(*args, **kwargs)

            class Connection(super(Psycopg2cffiDriver, self)._create_connection(mod, 'readonly')):
                def __init__(self, dsn, **kwargs):
                    super(Connection, self).__init__(dsn, **kwargs)
                    self.cursor_factory = Cursor
                    self.readonly = None

            mod.RSPsycopg2cffiConnection = Connection

        return mod.RSPsycopg2cffiConnection

    def _get_extension_module(self):
        from psycopg2cffi import extensions # pylint:disable=no-name-in-module
        return extensions

    # as of psycopg2cffi 2.8.1 connection has no '.info' attribute.

    def debug_connection(self, conn, *extra): # pragma: no cover
        print(conn, *extra)

    def connection_may_need_rollback(self, conn):
        return conn.status != self.STATUS_READY

    def connection_may_need_commit(self, conn):
        if conn.readonly:
            return False
        return self.connection_may_need_rollback(conn)

    def sync_status_after_hidden_commit(self, conn):
        conn.status = self.STATUS_READY



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
sqlite3 dialect.

There are a number of variations, depending what is supported by the
version of sqlite.

Sqlite supports ordered parameters with ``?`` and named parameters with ``:name``.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from sqlite3 import sqlite_version_info as sq3_version

from ..sql import DefaultDialect
from ..sql import Compiler as DefaultCompiler

from ..sql import OID
from ..sql import TID
from ..sql import BinaryString
from ..sql import State
from ..sql import Boolean

# Can sqlite3 enforce foreign key constraints? Note that you must
# enable this with 'PRAGMA foreign_keys=on'
SQ3_CAN_ENFORCE_FOREIGN_KEYS = sq3_version >= (3, 6, 19) # 2009-10-14

SQ3_SUPPORTS_WAL = sq3_version >= (3, 7)
SQ3_SUPPORTS_LENGTH_ON_BLOB = sq3_version >= (3, 7, 6)
SQ3_SUPPORTS_MULTIPLE_VALUES_INSERTS = sq3_version >= (3, 7, 11)
# Can multiple different connections share an in-memory database?
SQ3_SHARED_CACHE_SUPPORTS_MEMORY = sq3_version >= (3, 7, 13) # 2012-06-11

SQ3_SUPPORTS_CTE = sq3_version >= (3, 8, 3) # 2014-02-03

SQ3_SUPPORTS_PAREN_UPDATE = sq3_version >= (3, 15) # 2016-10-14
# Will it parse TRUE and FALSE as synonyms for 1/0 in queries?
SQ3_SUPPORTS_BOOL_KW = sq3_version >= (3, 23) # 2018-04-02
SQ3_SUPPORTS_UPSERT = sq3_version >= (3, 24) # 2018-06-04
SQ3_SUPPORTS_WINDOW = sq3_version >= (3, 25) # 2018-09-15



class _Sqlite3UpsertCompiler(DefaultCompiler):
    # We inherit (most of) the default upsert syntax (sqlite matches postgres
    # in that area).

    def can_prepare(self):
        # sqlite3 has no notion of prepared statements.
        return False

    def emit_identifier(self, identifier, quoted=False):
        if identifier == "transaction":
            # Sigh. This has to be quoted for some reason.
            quoted = True
        super(_Sqlite3UpsertCompiler, self).emit_identifier(identifier, quoted)

    if not SQ3_SUPPORTS_BOOL_KW:
        def visit_boolean_literal_expression(self, value):
            assert isinstance(value, bool)
            self.emit('1' if value else '0')

    def _placeholder(self, key):
        if key == '?':
            return key
        return ':%s' % (key,)

    def visit_ordered_bind_param(self, bind_param):
        self.placeholders[bind_param] = '?'
        self.emit('?')

    def visit_no_values(self):
        # Called following a VALUES statement when we have nothing.
        # But sqlite doesn't like (), there must be at least one
        # column
        self.emit("(NULL)")

    def visit_upsert_after_select(self, select):
        if select.is_unconstrained():
            # "The parser might not be able to tell if the "ON" keyword is
            # introducing the UPSERT or if it is the ON clause of a join.
            # To work around this, the SELECT statement should always
            # include a WHERE clause, even if that WHERE clause is just
            # ``WHERE true``."
            self.emit(' WHERE true ')

class _Sqlite3NoUpsertCompiler(_Sqlite3UpsertCompiler):
    # we have to disable the upsert syntax, we're limited to the old
    # syntax. The old 'INSERT OR REPLACE' syntax is supported from
    # 3.0.0 forward. It's not as flexible as the true upsert: for
    # example, you can't specify the type of conflict, nor can you use
    # a WHERE clause or specify the values to use in the update. It
    # also doesn't increment the cursor's change counter for replaced
    # rows.

    def emit_keyword_upsert(self):
        self.emit_keyword("INSERT OR REPLACE INTO")

    def visit_upsert_conflict_column(self, column):
        pass

    def visit_upsert_conflict_update(self, update):
        pass

    def visit_upsert_excluded_column(self, column):
        pass

    def visit_upsert_after_select(self, select):
        "No need to do anything"

SqliteCompiler = _Sqlite3UpsertCompiler if SQ3_SUPPORTS_UPSERT else _Sqlite3NoUpsertCompiler

class Sqlite3Dialect(DefaultDialect):
    datatype_map = {
        OID: 'INTEGER',
        TID: 'INTEGER',
        BinaryString: 'BLOB',
        State: 'BLOB',
        Boolean: 'INTEGER',
    }

    STMT_TRUNCATE = 'DELETE FROM'
    STMT_IF_NOT_EXISTS = 'IF NOT EXISTS'

    CONSTRAINT_AUTO_INCREMENT = 'AUTOINCREMENT'

    def compiler_class(self):
        return SqliteCompiler



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
psycopg2 IDBDriver implementations.
"""

from __future__ import absolute_import
from __future__ import print_function

from zope.interface import implementer

from relstorage.adapters.drivers import GeventConnectionMixin
from relstorage.adapters.drivers import GeventDriverMixin

from ...interfaces import IDBDriver
from ...interfaces import IDBDriverSupportsCritical
from ...drivers import MemoryViewBlobDriverMixin

from . import AbstractPostgreSQLDriver
from ._lobject import LobConnectionMixin

__all__ = [
    'Psycopg2Driver',
    'GeventPsycopg2Driver',
]

class WaitCallbackStateError(ImportError):

    def __init__(self, driver_name, wait_callback_installed, need_wait_callback):
        ImportError.__init__(
            self,
            "The driver %r %s a wait callback but %s wait callback is installed" % (
                driver_name,
                "needs" if need_wait_callback else "cannot have",
                "a" if wait_callback_installed else "no"
            )
        )

@implementer(IDBDriver)
class Psycopg2Driver(MemoryViewBlobDriverMixin,
                     AbstractPostgreSQLDriver):
    __name__ = 'psycopg2'
    MODULE_NAME = __name__

    PRIORITY = 1
    PRIORITY_PYPY = 2

    def __init__(self):
        super(Psycopg2Driver, self).__init__()

        psycopg2 = self.get_driver_module()

        # pylint:disable=no-member

        self.Binary = psycopg2.Binary
        self.connect = self._create_connection(psycopg2)

        # extensions
        self.ISOLATION_LEVEL_READ_COMMITTED = psycopg2.extensions.ISOLATION_LEVEL_READ_COMMITTED
        self.ISOLATION_LEVEL_SERIALIZABLE = psycopg2.extensions.ISOLATION_LEVEL_SERIALIZABLE
        self.ISOLATION_LEVEL_REPEATABLE_READ = psycopg2.extensions.ISOLATION_LEVEL_REPEATABLE_READ
        self.STATUS_READY = psycopg2.extensions.STATUS_READY

        self.TS_ACTIVE = psycopg2.extensions.TRANSACTION_STATUS_ACTIVE
        self.TS_INTRANS = psycopg2.extensions.TRANSACTION_STATUS_INTRANS
        self.TS_INERROR = psycopg2.extensions.TRANSACTION_STATUS_INERROR

        self.TS_NEEDS_COMMIT = (self.TS_ACTIVE, self.TS_INTRANS)
        self.TS_NOT_NEEDROLLBACK = psycopg2.extensions.TRANSACTION_STATUS_IDLE

    def _create_connection(self, mod, *extra_slots):
        if getattr(mod, 'RSPsycopg2Connection', self) is self:
            class RSPsycopg2Connection(mod.extensions.connection):
                # The replica attribute holds the name of the replica this
                # connection is bound to.
                __slots__ = ('replica',) + extra_slots

            mod.RSPsycopg2Connection = RSPsycopg2Connection

        return mod.RSPsycopg2Connection

    def _get_extension_module(self):
        # Subclasses should override this method if they use a different
        # DB-API module.
        from psycopg2 import extensions
        return extensions

    _WANT_WAIT_CALLBACK = False

    def _check_wait_callback(self):
        # Subclasses may override this method if they have no concept of
        # this.
        extensions = self._get_extension_module()
        callback = extensions.get_wait_callback()
        # Truth table:
        # callback is not None | Need callback    | pass?
        # True                 | True             | Pass
        # False                | False            | Pass
        # True                 | False            | Fail
        # False                | True             | Fail
        #
        # This is XNOR, exclusive nor, or iff. Note how it passes
        # only when both conditions are the same.

        callback_not_none = bool(callback is not None)
        need_callback = bool(self._WANT_WAIT_CALLBACK)
        if callback_not_none is not need_callback:
            raise WaitCallbackStateError(self.__name__, callback_not_none, need_callback)

    def get_driver_module(self):
        self._check_wait_callback()
        return super(Psycopg2Driver, self).get_driver_module()

    def connect_with_isolation(self, dsn,
                               isolation=None,
                               read_only=False,
                               deferrable=False,
                               application_name=None):
        # Even though the documentation claims that "Any other
        # connection parameter supported by the client library/server
        # can be passed either in the connection string or as a
        # keyword," it doesn't allow application_name as a kwarg. So we have to
        # resort to changing the dsn.
        if application_name and 'application_name' not in dsn:
            dsn = "%s application_name='%s'" % (dsn, application_name)
        conn = self.connect(dsn)
        assert not conn.autocommit
        if isolation or deferrable or read_only:
            conn.set_session(isolation_level=isolation, readonly=read_only,
                             deferrable=deferrable)

        return conn

    def cursor(self, conn, server_side=False):
        if server_side:
            cursor = conn.cursor(name=str(id(conn)))
            cursor.arraysize = self.cursor_arraysize
            cursor.itersize = self.cursor_arraysize
        else:
            cursor = super(Psycopg2Driver, self).cursor(conn)
        return cursor

    def debug_connection(self, conn, *extra): # pragma: no cover
        print(conn,
              'ts', conn.info.transaction_status,
              's', conn.status, 'tss', conn.info.status,
              'readonly', conn.readonly,
              "needs commit", self.connection_may_need_commit(conn),
              "needs rollback", self.connection_may_need_rollback(conn),
              *extra)

    def connection_may_need_rollback(self, conn):
        # If we've immediately executed a 'BEGIN' command,
        # the connection will report itself in a transaction, but
        # unless we've actually executed some sort of statement the
        # database will still know we're not and could issue a warning.
        result = conn.info.transaction_status != self.TS_NOT_NEEDROLLBACK or conn.readonly
        return result

    def connection_may_need_commit(self, conn):
        if conn.readonly:
            return False
        return conn.info.transaction_status in self.TS_NEEDS_COMMIT

    def sync_status_after_hidden_commit(self, conn):
        """
        If you execute a multiple statement or function that commits,
        *you must immediately* call this method.
        """
        # Sadly we can't do anything except commit. The .status
        # variable is untouchable
        self.commit(conn)


@implementer(IDBDriverSupportsCritical)
class GeventPsycopg2Driver(GeventDriverMixin, Psycopg2Driver):
    __name__ = 'gevent ' + Psycopg2Driver.MODULE_NAME

    _GEVENT_CAPABLE = True
    _GEVENT_NEEDS_SOCKET_PATCH = False

    supports_copy = False

    def _create_connection(self, mod, *extra_slots):
        if getattr(mod, 'RSGeventPsycopg2Connection', self) is self:
            Base = super(GeventPsycopg2Driver, self)._create_connection(mod, *extra_slots)

            class RSGeventPsycopg2Connection(GeventConnectionMixin,
                                             LobConnectionMixin,
                                             Base):
                RSDriverBinary = self.Binary
                _in_critical_phase = False

                def enter_critical_phase_until_transaction_end(self):
                    self._in_critical_phase = True

                def exit_critical_phase(self):
                    self._in_critical_phase = False

                def is_in_critical_phase(self):
                    return self._in_critical_phase

                def commit(self):
                    try:
                        super(RSGeventPsycopg2Connection, self).commit()
                    finally:
                        self._in_critical_phase = False

                def rollback(self):
                    try:
                        super(RSGeventPsycopg2Connection, self).rollback()
                    finally:
                        self._in_critical_phase = False

            mod.RSGeventPsycopg2Connection = RSGeventPsycopg2Connection

        return mod.RSGeventPsycopg2Connection

    def get_driver_module(self):
        # Make sure we can use gevent; if we can't the ImportError
        # will prevent this driver from being used.
        # TODO: Unify this better with inheritance.
        __import__('gevent')
        return super(GeventPsycopg2Driver, self).get_driver_module()

    _WANT_WAIT_CALLBACK = True

    ###
    # Critical sections
    ###

    def enter_critical_phase_until_transaction_end(self, connection, cursor):
        connection.enter_critical_phase_until_transaction_end()

    def is_in_critical_phase(self, connection, cursor):
        return connection.is_in_critical_phase()

    def exit_critical_phase(self, connection, cursor):
        connection.exit_critical_phase()

    # Because we have no insight into each individual query that's
    # run, and one of our queries commits in a single multi-statement,
    # if we're in critical phase, we need to commit or rollback to
    # revert that....except, the only time we do a multi-statement
    # that executes a commit we immediately turn around and call
    # ``sync_status_after_commit``, which in turn always calls
    # ``commit()``. So that's the simplest way. We've made this more
    # explicit with the
    # ``execute_multiple_statement_with_hidden_commit`` API. This
    # automatically ends a critical section.
    #
    # Previously, we briefly thought it necessary to override the
    # ``may_need_rollback`` and ``may_need_commit`` APIs.


class _GeventPsycopg2WaitCallback(object):

    def __init__(self):
        from gevent.socket import wait_read
        from gevent.socket import wait_write
        self.wait_read = wait_read
        self.wait_write = wait_write
        # pylint:disable=import-error,no-name-in-module
        from psycopg2.extensions import POLL_OK
        from psycopg2.extensions import POLL_WRITE
        from psycopg2.extensions import POLL_READ
        self.poll_ok = POLL_OK
        self.poll_write = POLL_WRITE
        self.poll_read = POLL_READ

    def __call__(self, conn):
        allow_switch = not conn._in_critical_phase
        # Even though ``conn.poll()`` would appear to do this, we need
        # to check at this level. Otherwise we can get strange hangs,
        # and one connection can monopolize the whole thing. Note that
        # this only applies when allow_switch is true.
        poll = True
        while 1:
            state = conn.poll()
            if state == self.poll_ok:
                return

            if state == self.poll_read:
                conn.gevent_generic_wait_read(allow_switch, poll)
            else:
                conn.gevent_generic_wait_write(allow_switch, poll)


def _gevent_did_patch(_event):
    try:
        from psycopg2.extensions import set_wait_callback
    except ImportError:
        pass
    else:
        set_wait_callback(_GeventPsycopg2WaitCallback())



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import

from relstorage._compat import PYPY


class Options(object):
    """Options for configuring and tuning RelStorage.

    These parameters can be provided as keyword options in the :class:`.RelStorage`
    constructor.  For example:

        storage = RelStorage(adapter, pack_gc=True, pack_prepack_only=True)

    Alternatively, the RelStorage constructor accepts an options
    parameter, which should be an Options instance.

    .. versionchanged:: 3.0
       No longer accepts the ``poll_interval`` option.

    .. versionchanged:: 2.0b2
       The ``poll_interval`` option is now ignored and raises a warning. It is always
       effectively 0.

    .. versionchanged:: 2.0b7
       The ``cache_delta_size_limit`` grew to 20000 on CPython.

    .. versionchanged:: 3.0a7
       Add the ``blob_cache_size_check_external`` option, defaulting to False.

       Change the default for ``shared_blob_dir`` to False.
    """

    #: The adapter factory configuration.
    adapter = None

    #: The name
    name = None
    #: Read only?
    read_only = False
    #: Preserve history?
    keep_history = True


    ###
    # Blob configuration
    ###

    #: directory to hold blobs
    blob_dir = None
    #: Is the blob directory shared?
    shared_blob_dir = False
    #: How much local blob data to keep
    blob_cache_size = None
    #: Percentage of ``blob_cache_size``; when any individual
    #: storage instance has written this many bytes of data into
    #: the blob cache, launch a cleanup process.
    blob_cache_size_check = 10
    #: Run the cache size check in an external process?
    blob_cache_size_check_external = False
    #: The size to break blobs into for storage.
    #: Only applies to some databases.
    blob_chunk_size = 1 << 20


    #: File containing replica info
    replica_conf = None
    #: File containing read-only replica info
    ro_replica_conf = None
    #: ?
    replica_timeout = 600.0
    #: Specifies what to do when a database connection is stale.
    revert_when_stale = False
    #: Perform a GC when packing
    pack_gc = True
    #: Only prepack
    pack_prepack_only = False
    #: Skip prepack
    pack_skip_prepack = False
    #: Length of time to hold a lock.
    pack_batch_timeout = 1.0

    #: List of memcache servers
    cache_servers = ()  # ['127.0.0.1:11211']
    #: Module to wrap a memcache connection with.
    cache_module_name = 'relstorage.pylibmc_wrapper'
    #: Database-specific prefix key
    cache_prefix = ''
    #: How much memory to use for the pickle cache
    cache_local_mb = 10
    #: The largest pickle to hold in the pickle cache
    cache_local_object_max = 16384
    #: How to compress local pickles
    cache_local_compression = 'none'
    #: Directory holding persistent cache files
    cache_local_dir = None
    #: Switch checkpoints after this many writes
    cache_delta_size_limit = 100000 if not PYPY else 50000

    #: How long to wait for a commit lock, in seconds.
    commit_lock_timeout = 30
    #: Lock ID for Oracle
    commit_lock_id = 0

    #: Automatically create the schema if needed
    create_schema = True
    #: Which database driver to use
    driver = 'auto'

    # Deprecated things
    #: How many persistent cache files to keep
    cache_local_dir_count = None
    #: How many persistent cache files to read
    cache_local_dir_read_count = None
    #: How big a cache file can be
    cache_local_dir_write_max_size = None
    #: Compress the cache files?
    cache_local_dir_compress = None
    #: How long to pause if we can't get the lock.
    #:
    #: ..versionchanged:: 3.0a5
    #:    Deprecated and unused.
    pack_commit_busy_delay = 5.0

    # We'll emit a warning for each option seen in here.
    _deprecated_options = (
        'cache_local_dir_count',
        'cache_local_dir_read_count',
        'cache_local_dir_write_max_size',
        'cache_local_dir_compress',
        'pack_commit_busy_delay',
    )


    def __init__(self, **kwoptions):
        for key, value in kwoptions.items():
            if not hasattr(self, key):
                raise TypeError("Unknown parameter: %s (Known: %s)" % (
                    key,
                    self.valid_option_names()
                ))
            if key in self._deprecated_options:
                import warnings
                warnings.warn(
                    "The option %s is deprecated and ignored." % (key,),
                    FutureWarning
                )

            setattr(self, key, value)

    @classmethod
    def copy_valid_options(cls, other_options):
        """
        Produce a new options featuring only the valid settings from
        *other_options*.
        """
        option_dict = {}
        for key in cls.valid_option_names():
            value = getattr(other_options, key, None)
            if value is not None:
                option_dict[key] = value
        return cls(**option_dict)

    @classmethod
    def valid_option_names(cls):
        return sorted(
            x
            for x in vars(cls)
            if not callable(getattr(cls, x)) and not x.startswith('_')
        )

    def __repr__(self):
        opts = []
        for k, v in sorted(self.__dict__.items()):
            opt = '%s=%r' % (k, v)
            opts.append(opt)
        opts = ', '.join(opts)
        return 'relstorage.options.Options(%s)' % (opts,)

    def __eq__(self, other):
        if not isinstance(other, Options):
            return NotImplemented
        return all(getattr(self, key) == getattr(other, key)
                   for key in self.valid_option_names())

    def copy(self, **kw):
        """
        Produce a copy of these options, with keyword arguments overriding.
        """
        options = dict(self.__dict__)
        options.update(kw)
        return self.__class__(**options)

    new_instance = copy



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""PostgreSQL adapter for RelStorage."""
from __future__ import absolute_import
from __future__ import print_function

from relstorage.adapters.postgresql.adapter import PostgreSQLAdapter


assert PostgreSQLAdapter


def debug_locks(cursor, me_only=False, exclusive_only=False): # pragma: no cover
    if me_only:
        conn = cursor.connection
        try:
            pid = conn.info.backend_pid
        except AttributeError:
            # connection.info was added to psycopg2 in 2.8.0
            cursor.execute("SELECT pg_backend_pid()")
            pid = cursor.fetchall()[0][0]
        params = (pid,)
        pid_clause = ' AND a.pid = %s '
    else:
        pid_clause = ''
        params = ()
    excl_clause = ''
    if exclusive_only:
        excl_clause = """ AND l.mode like '%%Exclu%%' """

    stmt = """
    SELECT
         a.datname,
         l.relation::regclass,
         l.locktype,
         l.mode,
         l.pid,
         l.virtualxid, l.transactionid,
         l.objid, l.objsubid,
         l.virtualtransaction,
         l.granted
    FROM pg_stat_activity a
    JOIN pg_locks l ON l.pid = a.pid
    WHERE l.locktype <> 'virtualxid'
    {pid_clause}
    {excl_clause}
    ORDER BY a.pid
    """.format(
        pid_clause=pid_clause,
        excl_clause=excl_clause
    )
    cursor.execute(stmt, params)
    return cursor

def debug_my_locks(cursor, exclusive_only=False): # pragma: no cover
    return debug_locks(cursor, me_only=True, exclusive_only=exclusive_only)

def print_size_report(cursor): # pragma: no cover
    extra_query = ''
    query = """
    SELECT table_name,
        pg_size_pretty(total_bytes) AS total,
        pg_size_pretty(index_bytes) AS INDEX,
        pg_size_pretty(toast_bytes) AS toast,
        pg_size_pretty(table_bytes) AS TABLE
    FROM (
    SELECT *, total_bytes-index_bytes-COALESCE(toast_bytes,0) AS table_bytes FROM (
    SELECT c.oid,nspname AS table_schema, relname AS TABLE_NAME
          , c.reltuples AS row_estimate
          , pg_total_relation_size(c.oid) AS total_bytes
          , pg_indexes_size(c.oid) AS index_bytes
          , pg_total_relation_size(reltoastrelid) AS toast_bytes
      FROM pg_class c
      LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
      WHERE relkind = 'r'
        ) a
    ) a
    WHERE (table_name NOT LIKE 'pg_%' and table_name not like 'abstract_%'
    )
    AND table_schema <> 'pg_catalog' and table_schema <> 'information_schema'
    ORDER BY total_bytes DESC
    """

    cursor.execute(query)
    keys = ['table_name', 'total', 'index', 'toast', 'table']
    rows = [dict(zip(keys, row)) for row in cursor]

    rows.insert(0, {k: k for k in keys})
    print()
    fmt = "| {table_name:35s} | {total:10s} | {index:10s} | {toast:10s} | {table:10s}"
    for row in rows:
        if not extra_query and row['total'] in ('72 kB', '32 kB', '24 kB', '16 kB', '8192 bytes'):
            continue
        print(fmt.format(
            **{k: v if v else '<null>' for k, v in row.items()}
        ))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2016 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
pg8000 IDBDriver implementations.
"""

from __future__ import absolute_import
from __future__ import print_function

from collections import deque

from zope.interface import implementer

from relstorage._compat import number_types

from ...interfaces import IDBDriver
from ...sql import Compiler

from . import AbstractPostgreSQLDriver
from . import PostgreSQLDialect
from ._lobject import LobConnectionMixin


__all__ = [
    'PG8000Driver',
]


class _tuple_deque(deque):

    def append(self, row): # pylint:disable=arguments-differ
        deque.append(self, tuple(row))

class PG8000Compiler(Compiler):

    def can_prepare(self):
        # Important: pg8000 1.10 - 1.13, at least, can't handle prepared
        # statements that take parameters but it doesn't need to because it
        # prepares every statement anyway. So you must have a backup that you use
        # for that driver.
        # https://github.com/mfenniak/pg8000/issues/132
        return False

class PG8000Dialect(PostgreSQLDialect):

    def compiler_class(self):
        return PG8000Compiler


@implementer(IDBDriver)
class PG8000Driver(AbstractPostgreSQLDriver):
    __name__ = 'pg8000'
    MODULE_NAME = __name__
    PRIORITY = 3
    PRIORITY_PYPY = 2

    _GEVENT_CAPABLE = True
    _GEVENT_NEEDS_SOCKET_PATCH = True

    dialect = PG8000Dialect()
    supports_multiple_statement_execute = False
    supports_64bit_unsigned_id = False

    def __init__(self):
        super(PG8000Driver, self).__init__()
        # XXX: global side-effect!
        self.driver_module.paramstyle = 'pyformat'

        # XXX: Testing. Can we remove?
        self.disconnected_exceptions += (AttributeError,)

        # Sadly, pg8000 raises ProgrammingError when it can't
        # obtain a lock, making it hard to distinguish. We'll let other
        # clauses catch such exceptions.
        self.illegal_operation_exceptions = ()
        Binary = self.Binary

        if getattr(self.driver_module, 'RSConnection', self) is self:
            class Cursor(self.driver_module.Cursor):
                def __init__(self, conn):
                    super(Cursor, self).__init__(conn)
                    # pylint:disable=access-member-before-definition
                    assert isinstance(self._cached_rows, deque)
                    # Make sure rows are tuples, not lists.
                    # BTrees don't like lists.
                    self._cached_rows = _tuple_deque()

                @property
                def connection(self):
                    # silence the warning it wants to generate:
                    # "UserWarning: DB-API extension cursor.connection used"
                    return self._c

                def copy_expert(self, sql, stream):
                    return self.execute(sql, stream=stream)

            class Connection(LobConnectionMixin,
                             self.driver_module.Connection):
                readonly = False
                RSDriverBinary = staticmethod(Binary)

                def __init__(self,
                             user, host='localhost',
                             unix_sock=None,
                             port=5432, database=None,
                             password=None, ssl_context=None,
                             timeout=None, application_name=None,
                             max_prepared_statements=1000,
                             tcp_keepalive=True):
                    # pylint:disable=useless-super-delegation

                    # Prior to pg8000 1.15.1, we had to do this
                    # because the super class requires all these
                    # arguments and didn't have defaults. In that release, it gained
                    # defaults, but our order had been wrong for some time. In
                    # 1.15.3 our bad ordering started causing TypeError.
                    # Pass these things in this way to be compatible with older versions too
                    # that didn't have defaults or extra parameters, but the names
                    # were the same...except for `ssl` vs `ssl_context`: the latter
                    # is only in versions that run on Python 3.
                    # We must retain support for older versions for Python 2.
                    #
                    # We could probably do better than this
                    kwargs = dict(
                        user=user, host=host, unix_sock=unix_sock,
                        port=port, database=database,
                        password=password, ssl_context=ssl_context,
                        timeout=timeout,
                        application_name=application_name,
                        max_prepared_statements=max_prepared_statements,
                        tcp_keepalive=tcp_keepalive
                    )
                    if str is bytes: # PY2
                        del kwargs['ssl_context']
                        kwargs['ssl'] = ssl_context
                    try:
                        super(Connection, self).__init__(**kwargs)
                    except TypeError:
                        # 1.16.0 dropped the ``max_prepared_statements`` argument.
                        del kwargs['max_prepared_statements']
                        super(Connection, self).__init__(**kwargs)

                def cursor(self):
                    return Cursor(self)

            self.driver_module.RSConnection = Connection

        self._connect = self.driver_module.RSConnection

    def connect(self, dsn, application_name=None): # pylint:disable=arguments-differ
        # Parse the DSN into parts to pass as keywords.
        # We don't do this psycopg2 because a real DSN supports more options than
        # we do and we don't want to limit it.
        kwds = {}
        parts = dsn.split(' ')
        for part in parts:
            key, value = part.split('=')
            value = value.strip("'\"")
            key = 'database' if key == 'dbname' else key
            value = int(value) if key == 'port' else value
            kwds[key] = value
        kwds['application_name'] = application_name
        conn = self._connect(**kwds)
        return conn

    ISOLATION_LEVEL_READ_COMMITTED = 'ISOLATION LEVEL READ COMMITTED'
    ISOLATION_LEVEL_SERIALIZABLE = 'ISOLATION LEVEL SERIALIZABLE'
    ISOLATION_LEVEL_REPEATABLE_READ = 'ISOLATION LEVEL REPEATABLE READ'

    def connect_with_isolation(self, dsn,
                               isolation=None,
                               read_only=False,
                               deferrable=False,
                               application_name=None):
        conn = self.connect(dsn, application_name=application_name)
        cursor = self.cursor(conn)
        # For the current transaction
        transaction_stmt = 'TRANSACTION %s %s %s' % (
            isolation,
            ', READ ONLY' if read_only else '',
            ', DEFERRABLE' if deferrable else ''
        )
        cursor.execute('SET ' + transaction_stmt)
        # For future transactions on this same connection.
        # NOTE: This will probably not play will with things like pgbouncer.
        # See http://initd.org/psycopg/docs/connection.html#connection.set_session
        cursor.execute('SET SESSION CHARACTERISTICS AS ' + transaction_stmt)
        conn.commit()
        conn.readonly = read_only
        return conn

    sql_compiler_class = PG8000Compiler

    def set_lock_timeout(self, cursor, timeout):
        # PG8000 needs a literal embedded in the string; prepared
        # statements can't be SET with a variable.
        assert isinstance(timeout, number_types)
        cursor.execute('SET lock_timeout = %s' % (timeout,))

    # TODO: Implement 'connection_may_need_commit' and 'connection_may_need_rollback'
    # based on the actual state of the transaction. In psycopg2, we have the .status
    # or .info to tell us; can we find something similar here?
    def connection_may_need_rollback(self, conn):
        # Got to rollback even if we are read only to get fresh view of the database.
        return True

    def connection_may_need_commit(self, conn):
        if conn.readonly:
            return False
        return conn.in_transaction



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
sqlite3 adapter for RelStorage.

General Design Notes
====================

Sqlite3 ``INTEGER PRIMARY KEY`` values are always 64-bit *signed*
integers. If not given, they default to an unused integer (typically
incrementing). If the ``AUTOINCREMENT`` keyword is provided, then they
prevent reuse of values even after they are deleted via an auxiliary
table, which adds overhead.

Regular tables are structured as a B*Tree with internal nodes
containing only the ROWID (a 64-bit integer) and all row content
stored in the leaf pages.

If a regular table has an ``INTEGER PRIMARY KEY`` column (it must be
exactly ``INTEGER``, not ``INT``) then that column becomes an alias
for the ROWID. It is enforced to be an actual integer (unlike all
other columnns that can store arbitrary data). Accesses by ROWID are
very efficient. That's a natural for history-free ``OBJECT_STATE`` and
history-preserving ``TRANSACTION`` tables.

If a regular table has a primary key of any other type or of multiple
columns, then it requires two BTrees: one to store the rows, and a
second ``UNIQUE`` index table to implement the primary key and map it
back to the ROWID of the data BTree.

It is possible to create tables ``WITHOUT ROWID``. These are normal
BTrees, with all the data for a row stored in the node where it is
found, whether leaf or interior. For multi-column primary keys, this
can be very attractive because it avoids duplicate storage of the
primary key, and allows direct access to the rest of the row given the
primary key instead of requiring redirection throgh the index.

However, because these tables store data inline, it is not recommended
to use them if the row could be arbitrarily large (more than 200
bytes) because of the effect that has on the BTree and number of pages
needed to traverse it. That makes it a poor fit for history-preserving
``OBJECT_STATE``.

References:

    - https://www.sqlite.org/lang_createtable.html#rowid

    - https://www.sqlite.org/withoutrowid.html

OIDs
====

Our ``object_state`` table will use a regular ``INTEGER PRIMARY KEY``
table. Our ``new_oids`` table COULD use one defined to
``AUTOINCREMENT``. However, that complicates copying transactions, the
implementation of :meth:`IOIDAllocator.set_min_oid`, and introduces
extra ``DELETE`` operations to ensure that the table doesn't grow
forever, so we handle the incrementing ourself.

We keep ``new_oids`` in a separate database file beside the primary
database file. This is because *any* write to a database causes an
exclusive write lock to be taken. Allocating new oids can happen at
essentially any time, even outside of the normal two-phase commit
sequence (thanks to Connection.add()) and we cannot allow the main
database to be locked like that.

Because OIDs are signed 64-bit integers, they cannot store values
larger than 2**63 - 1 (around nine quintillion)

TIDs
====

Because TIDs are signed 64-bit integers, they cannot store values
larger than 2**63 - 1, or nine quintillion transactions. Since TID is
related to the current time, that suffices for values until the year 5908.

Locks
=====

Sqlite3 only supports database-wide write locks. For details on when
and how they are taken and managed, see connmanager.py and locker.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.adapters.tests import test_txncontrol


class TestPostgreSQLTransactionControl(test_txncontrol.TestTransactionControl):

    def _getClass(self):
        from ..txncontrol import PostgreSQLTransactionControl
        return PostgreSQLTransactionControl

    def _get_hf_tid_query(self):
        return 'EXECUTE get_latest_tid'

    _get_hp_tid_query = _get_hf_tid_query



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from relstorage.tests import TestCase
from relstorage.tests import MockCursor
from relstorage._util import consume


class RowBatcherTests(TestCase):

    def getClass(self):
        from relstorage.adapters.batch import RowBatcher
        return RowBatcher

    def test_delete_defer(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.delete_from("mytable", id=2)
        self.assertEqual(cursor.executed, [])
        self.assertEqual(batcher.rows_added, 1)
        self.assertEqual(batcher.size_added, 0)
        self.assertEqual(batcher.total_rows_inserted, 0)
        self.assertEqual(batcher.total_rows_deleted, 0)
        self.assertEqual(batcher.total_size_inserted, 0)
        self.assertEqual(dict(batcher.deletes),
                         {('mytable', ('id',)): set([(2,)])})

    def test_delete_multiple_column(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.delete_from("mytable", id=2, tid=10)
        self.assertEqual(cursor.executed, [])
        self.assertEqual(batcher.rows_added, 1)
        self.assertEqual(batcher.size_added, 0)
        self.assertEqual(dict(batcher.deletes),
                         {('mytable', ('id', 'tid')): set([(2, 10)])})

    IN_ROWS_FLATTENED = False
    delete_auto_flush = 'DELETE FROM mytable WHERE id IN (%s,%s)'

    def _in(self, *params):
        params = sorted(params)
        if self.IN_ROWS_FLATTENED:
            l = list(params)
            return (l,)
        return tuple(params)

    def test_delete_auto_flush(self):
        cursor = MockCursor()
        cursor.sort_sequence_params = True
        batcher = self.getClass()(cursor, 2)
        batcher.sorted_deletes = True
        batcher.delete_from("mytable", id=2)
        batcher.delete_from("mytable", id=1)
        self.assertEqual(
            cursor.executed,
            [
                (self.delete_auto_flush,
                 self._in(1, 2)
                )
            ])
        self.assertEqual(batcher.rows_added, 0)
        self.assertEqual(batcher.size_added, 0)
        self.assertEqual(batcher.deletes, {})
        self.assertEqual(batcher.total_rows_inserted, 0)
        self.assertEqual(batcher.total_rows_deleted, 2)
        self.assertEqual(batcher.total_size_inserted, 0)

    def test_insert_defer(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=3,
        )
        self.assertEqual(cursor.executed, [])
        self.assertEqual(batcher.rows_added, 1)
        self.assertEqual(batcher.size_added, 3)
        self.assertEqual(batcher.inserts, {
            ('INSERT', 'mytable (id, name)', '%s, id || %s', ''): {1: (1, 'a')}
        })
        self.assertEqual(batcher.total_rows_inserted, 0)
        self.assertEqual(batcher.total_rows_deleted, 0)
        self.assertEqual(batcher.total_size_inserted, 0)

    def test_insert_defer_multi_table(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=3,
        )
        batcher.insert_into(
            "othertable (name)",
            "?",
            ('a'),
            rowkey=1,
            size=1,
        )

        self.assertEqual(cursor.executed, [])
        self.assertEqual(batcher.rows_added, 2)
        self.assertEqual(batcher.size_added, 4)
        self.assertEqual(dict(batcher.inserts), {
            ('INSERT', 'mytable (id, name)', '%s, id || %s', ''): {1: (1, 'a')},
            ('INSERT', 'othertable (name)', '?', ''): {1: ('a')},
        })
        self.assertEqual(batcher.total_rows_inserted, 0)
        self.assertEqual(batcher.total_rows_deleted, 0)
        self.assertEqual(batcher.total_size_inserted, 0)

    def test_insert_replace(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=3,
            command='REPLACE',
        )
        self.assertEqual(cursor.executed, [])
        self.assertEqual(batcher.rows_added, 1)
        self.assertEqual(batcher.size_added, 3)
        self.assertEqual(batcher.inserts, {
            ('REPLACE', 'mytable (id, name)', '%s, id || %s', ''): {1: (1, 'a')}
        })

    def test_insert_duplicate(self):
        # A second insert on the same rowkey replaces the first insert.
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=3,
            )
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'b'),
            rowkey=1,
            size=3,
            )
        self.assertEqual(cursor.executed, [])
        self.assertEqual(batcher.rows_added, 2)
        self.assertEqual(batcher.size_added, 6)
        self.assertEqual(batcher.inserts, {
            ('INSERT', 'mytable (id, name)', '%s, id || %s', ''): {1: (1, 'b')}
        })

    def test_insert_auto_flush(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.size_limit = 10
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=5,
            )
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (2, 'B'),
            rowkey=2,
            size=5,
            )
        self.assertEqual(
            cursor.executed,
            [(
                'INSERT INTO mytable (id, name) VALUES\n'
                '(%s, id || %s),\n'
                '(%s, id || %s)\n',
                (1, 'a', 2, 'B'))
            ])
        self.assertEqual(batcher.rows_added, 0)
        self.assertEqual(batcher.size_added, 0)
        self.assertEqual(batcher.inserts, {})
        self.assertEqual(batcher.total_rows_inserted, 2)
        self.assertEqual(batcher.total_rows_deleted, 0)
        self.assertEqual(batcher.total_size_inserted, 10)


    def test_insert_auto_flush_multi_table(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        batcher.size_limit = 10
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=5,
            )
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (2, 'B'),
            rowkey=2,
            size=5,
            )
        self.assertEqual(
            cursor.executed,
            [(
                'INSERT INTO mytable (id, name) VALUES\n'
                '(%s, id || %s),\n'
                '(%s, id || %s)\n',
                (1, 'a', 2, 'B'))
            ])
        self.assertEqual(batcher.rows_added, 0)
        self.assertEqual(batcher.size_added, 0)
        self.assertEqual(batcher.inserts, {})
        self.assertEqual(batcher.total_rows_inserted, 2)
        self.assertEqual(batcher.total_rows_deleted, 0)
        self.assertEqual(batcher.total_size_inserted, 10)

    flush_delete_one = 'DELETE FROM mytable WHERE id IN (?)'

    def test_flush(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor, delete_placeholder="?")
        # Make sure we preserve order in multi-column
        batcher.sorted_deletes = True
        batcher.delete_from("mytable", id=1)
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=5,
            )
        batcher.delete_from("mytable", id=1, key='abc')
        batcher.delete_from("mytable", id=2, key='def')
        batcher.flush()
        self.assertEqual(cursor.executed, [
            (self.flush_delete_one,
             self._in(1)),
            ('DELETE FROM mytable WHERE (id=? AND key=?) OR (id=? AND key=?)',
             (1, 'abc', 2, 'def')),
            ('INSERT INTO mytable (id, name) VALUES\n(%s, id || %s)\n',
             (1, 'a')),
        ])

    select_one = 'SELECT zoid,tid FROM object_state WHERE oids IN (%s)'

    def test_select_one(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor)
        consume(batcher.select_from(('zoid', 'tid'), 'object_state', oids=(1,)))
        self.assertEqual(cursor.executed, [
            (self.select_one,
             self._in(1,))
        ])

    select_multiple_one_batch = 'SELECT zoid,tid FROM object_state WHERE oids IN (%s,%s,%s,%s)'

    def test_select_multiple_one_batch(self):
        cursor = MockCursor()
        cursor.sort_sequence_params = True
        batcher = self.getClass()(cursor)
        list(batcher.select_from(('zoid', 'tid'), 'object_state',
                                 oids=(1, 2, 3, 4)))
        self.assertEqual(cursor.executed, [
            (self.select_multiple_one_batch,
             self._in(1, 2, 3, 4))
        ])

    select_multiple_many_batch = 'SELECT zoid,tid FROM object_state WHERE oids IN (%s,%s)'

    def test_select_multiple_many_batch(self, batch_limit_attr='row_limit'):
        cursor = MockCursor()
        cursor.sort_sequence_params = True
        cursor.many_results = [
            [(1, 1)],
            [(3, 1)],
            []
        ]
        batcher = self.getClass()(cursor)
        setattr(batcher, batch_limit_attr, 2)
        rows = batcher.select_from(('zoid', 'tid'), 'object_state',
                                   oids=iter((1, 2, 3, 4, 5)))
        rows = list(rows)

        self.assertEqual(cursor.executed, [
            (self.select_multiple_many_batch,
             self._in(1, 2)),
            (self.select_multiple_many_batch,
             self._in(3, 4)),
            (self.select_one,
             self._in(5)),
        ])

        self.assertEqual(rows, [
            (1, 1),
            (3, 1)
        ])

    def test_select_multiple_many_batch_bind_limit(self):
        self.test_select_multiple_many_batch(batch_limit_attr='bind_limit')


    def test_select_from_timeout(self):
        from relstorage.tests import mock
        from relstorage.adapters.interfaces import AggregateOperationTimeoutError
        cursor = MockCursor()
        cursor.sort_sequence_params = True
        cursor.many_results = [
            [(1, 1)],
            [(2, 1)],
            [(3, 1)],
            []
        ]
        batcher = self.getClass()(cursor)
        batcher.bind_limit = 1
        batcher.perf_counter = mock.Mock()
        # These will be the time values returned from perf_counter()
        batcher.perf_counter.side_effect = (
            12345, # Begin
            12346, # First batch
            12347, # Second batch
        )

        gener = batcher.select_from(('zoid', 'tid',), 'object_state',
                                    timeout=2,
                                    oids=[1, 2, 3, 4, 5])
        rows = []
        with self.assertRaises(AggregateOperationTimeoutError):
            for row in gener:
                rows.append(row)

        # We ran exactly twice before the perf_counter exceeded the timeout.
        self.assertEqual(rows, [
            (1, 1),
            (2, 1),
        ])




class OracleRowBatcherTests(TestCase):

    def getClass(self):
        from relstorage.adapters.oracle.batch import OracleRowBatcher
        return OracleRowBatcher

    def test_insert_one_row(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor, {})
        batcher.insert_into(
            "mytable (id, name)",
            "%s, id || %s",
            (1, 'a'),
            rowkey=1,
            size=3,
            )
        self.assertEqual(cursor.executed, [])
        batcher.flush()
        self.assertEqual(cursor.executed, [
            ('INSERT INTO mytable (id, name) VALUES (%s, id || %s)', (1, 'a')),
            ])

    def test_insert_two_rows(self):
        cursor = MockCursor()
        batcher = self.getClass()(cursor, {})
        batcher.insert_into(
            "mytable (id, name)",
            ":id, :id || :name",
            {'id': 1, 'name': 'a'},
            rowkey=1,
            size=3,
            )
        batcher.insert_into(
            "mytable (id, name)",
            ":id, :id || :name",
            {'id': 2, 'name': 'b'},
            rowkey=2,
            size=3,
            )
        self.assertEqual(cursor.executed, [])
        batcher.flush()
        self.assertEqual(
            cursor.executed,
            [(
                'INSERT ALL\n'
                'INTO mytable (id, name) VALUES (:id_0, :id_0 || :name_0)\n'
                'INTO mytable (id, name) VALUES (:id_1, :id_1 || :name_1)\n'
                'SELECT * FROM DUAL',
                {'id_0': 1, 'id_1': 2, 'name_1': 'b', 'name_0': 'a'})
            ])

    def test_insert_one_raw_row(self):
        class MockRawType(object):
            pass
        cursor = MockCursor()
        batcher = self.getClass()(cursor, {'rawdata': MockRawType})
        batcher.insert_into(
            "mytable (id, data)",
            ":id, :rawdata",
            {'id': 1, 'rawdata': 'xyz'},
            rowkey=1,
            size=3,
            )
        batcher.flush()
        self.assertEqual(cursor.executed, [
            ('INSERT INTO mytable (id, data) VALUES (:id, :rawdata)',
             {'id': 1, 'rawdata': 'xyz'})
        ])
        self.assertEqual(cursor.inputsizes, {'rawdata': MockRawType})

    def test_insert_two_raw_rows(self):
        class MockRawType(object):
            pass
        cursor = MockCursor()
        batcher = self.getClass()(cursor, {'rawdata': MockRawType})
        batcher.insert_into(
            "mytable (id, data)",
            ":id, :rawdata",
            {'id': 1, 'rawdata': 'xyz'},
            rowkey=1,
            size=3,
            )
        batcher.insert_into(
            "mytable (id, data)",
            ":id, :rawdata",
            {'id': 2, 'rawdata': 'abc'},
            rowkey=2,
            size=3,
            )
        batcher.flush()
        self.assertEqual(
            cursor.executed,
            [(
                'INSERT ALL\n'
                'INTO mytable (id, data) VALUES (:id_0, :rawdata_0)\n'
                'INTO mytable (id, data) VALUES (:id_1, :rawdata_1)\n'
                'SELECT * FROM DUAL',
                {'id_0': 1, 'id_1': 2, 'rawdata_0': 'xyz', 'rawdata_1': 'abc'})
            ])
        self.assertEqual(cursor.inputsizes, {
            'rawdata_0': MockRawType,
            'rawdata_1': MockRawType,
        })

class PostgreSQLRowBatcherTests(RowBatcherTests):

    def getClass(self):
        from relstorage.adapters.postgresql.batch import PostgreSQLRowBatcher
        return PostgreSQLRowBatcher

    IN_ROWS_FLATTENED = True

    delete_auto_flush = 'DELETE FROM mytable WHERE id = ANY (%s)'
    flush_delete_one = 'DELETE FROM mytable WHERE id = ANY (?)'
    select_one = 'SELECT zoid,tid FROM object_state WHERE oids = ANY (%s)'
    select_multiple_one_batch = 'SELECT zoid,tid FROM object_state WHERE oids = ANY (%s)'
    select_multiple_many_batch = 'SELECT zoid,tid FROM object_state WHERE oids = ANY (%s)'



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
A small abstraction layer for SQL queries.

Features:

    - Simple, readable syntax for writing queries.

    - Automatic switching between history-free and history-preserving
      schemas.

    - Support for automatically preparing statements (depending on the
      driver; some allow parameters, some do not, for example)

    - Always use bind parameters

    - Take care of minor database syntax issues.

This is inspired by the SQLAlchemy Core, but we don't use it because
we don't want to take a dependency that can conflict with applications
using RelStorage.
"""
# pylint:disable=too-many-lines

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .schema import Table
from .schema import TemporaryTable
from .schema import HistoryVariantTable
from .schema import Column
from .schema import ColumnResolvingProxy
from .schema import ColumnExpression
from .schema import View

from .dialect import DefaultDialect
from .dialect import Compiler

from .types import OID
from .types import TID
from .types import State
from .types import Boolean
from .types import BinaryString
from .types import Char

from .functions import func

it = ColumnResolvingProxy()

__all__ = [
    # Schema elements
    'Table',
    'Column',
    "TemporaryTable",
    "HistoryVariantTable",
    "View",

    # Query helpers
    'it',
    "ColumnExpression",

    # Dialect
    "DefaultDialect",
    "Compiler",

    # Types
    "OID",
    "TID",
    "State",
    "Boolean",
    "BinaryString",
    "Char",

    # Functions
    "func",

]



# -*- coding: utf-8 -*-
# distutils: language = c++
# cython: auto_pickle=False,embedsignature=True,always_allow_keywords=False,infer_types=True
"""
Python wrappers for the generational cache itself.

The cache controls object lifetime.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

cimport cython
from cython.operator cimport typeid
from cython.operator cimport dereference as deref
from cython.operator cimport preincrement as preincr
from cpython.buffer cimport PyBuffer_FillInfo
from cpython.bytes cimport PyBytes_AsString # Does NOT copy

from libcpp.pair cimport pair
from libcpp.cast cimport static_cast
from libcpp.cast cimport dynamic_cast
from libcpp.string cimport string

from relstorage.cache.c_cache cimport TID_t
from relstorage.cache.c_cache cimport OID_t
from relstorage.cache.c_cache cimport ICacheEntry
from relstorage.cache.c_cache cimport SVCacheEntry
from relstorage.cache.c_cache cimport MVCacheEntry
from relstorage.cache.c_cache cimport Cache
from relstorage.cache.c_cache cimport Generation
from relstorage.cache.c_cache cimport move
from relstorage.cache.c_cache cimport TempCacheFiller
from relstorage.cache.c_cache cimport ProposedCacheEntry

import sys
from relstorage.cache.interfaces import NoSuchGeneration
from relstorage.cache.interfaces import CacheConsistencyError

from relstorage._compat import iteroiditems

cdef extern from *:
    """
    #ifdef __clang__
    #pragma clang diagnostic push
    /* Cython generates lots of unreachable code diagnostics that flood the output */
    #pragma clang diagnostic ignored "-Wunreachable-code"
    /* As of Cython 3.0a6 and CPython 3.8 (at least) Cython generates
       deprecation warnings for tp_print */
    #pragma clang diagnostic ignored "-Wdeprecated-declarations"
    #endif

    template <typename T>
    T* array_new(int n) {
        return new T[n];
    }
    template <typename T>
    void array_delete(T* t) {
        delete[] t;
    }
    """
    T* array_new[T](int)
    void array_delete[T](T* t)

ctypedef const SVCacheEntry* SVCacheEntry_p
ctypedef const MVCacheEntry* MVCacheEntry_p
ctypedef const ICacheEntry*  ICacheEntry_p
ctypedef ICacheEntry** ICacheEntry_pp

cdef class CachedValue:
    """
    The base class for cached values.
    """

cdef inline CachedValue python_from_entry_p(ICacheEntry_p entry):
    cdef const SVCacheEntry* sve_p = dynamic_cast[SVCacheEntry_p](entry)
    cdef const MVCacheEntry* mve_p = dynamic_cast[MVCacheEntry_p](entry)
    if not sve_p and not mve_p:
        # Most likely, entry is null
        raise AssertionError("Invalid pointer cast")

    # Keeping in mind the semantics of our constructors
    # and Cython destructors, we'll let the class constructors handle that.
    cdef SingleValue sv
    if sve_p:
        sv = SingleValue.from_entry(sve_p)
        return sv

    return MultipleValues.from_entry(mve_p)


cdef inline CachedValue python_from_entry(const ICacheEntry& entry):
    return python_from_entry_p(&entry)

cdef inline object bytes_from_pickle(const SVCacheEntry_p entry):
    return entry.as_object()

ctypedef fused ConcreteCacheEntry:
    SVCacheEntry_p
    MVCacheEntry_p

cdef inline void release_entry(ConcreteCacheEntry* entry) except *:
    # Docs say to be very careful during __dealloc__
    if entry[0] and entry[0].Py_release():
        del entry[0]
    entry[0] = NULL


@cython.final
@cython.internal
cdef class SingleValue(CachedValue):
    cdef SVCacheEntry_p entry

    def __cinit__(self):
        self.entry = <SVCacheEntry_p>0

    @staticmethod
    cdef SingleValue from_entry(SVCacheEntry_p entry):
        cdef SingleValue sv = SingleValue.__new__(SingleValue)
        sv.entry = entry.Py_use[SVCacheEntry]()
        return sv

    def __dealloc__(self):
        release_entry(&self.entry)

    def sizeof(self):
        # At this writing, reports 88
        return sizeof(SVCacheEntry)

    def __iter__(self):
        return iter((
            bytes_from_pickle(self.entry),
            self.entry.tid()
        ))

    @property
    def frozen(self):
        return self.entry.frozen()

    @property
    def value(self):
        return self.state

    @property
    def key(self):
        return self.entry.key

    @property
    def frequency(self):
        return self.entry.frequency

    @property
    def state(self):
        return bytes_from_pickle(self.entry)

    @property
    def tid(self):
        return self.entry.tid()

    @property
    def max_tid(self):
        return self.entry.tid()

    @property
    def newest_value(self):
        return self

    @property
    def weight(self):
        return self.entry.weight()

    def __eq__(self, other):
        cdef SingleValue p
        if other is self:
            return True

        if isinstance(other, SingleValue):
            p = <SingleValue>other
            my_entry = self.entry
            other_entry = p.entry
            return my_entry == other_entry

        if isinstance(other, tuple):
            return len(other) == 2 and self.tid == other[1] and self.value == other[0]
        return NotImplemented

    def __getitem__(self, int i):
        if i == 0:
            return self.state()
        if i == 1:
            return self.entry.tid()
        raise IndexError

    def __repr__(self):
        return "%s(%r, %s, frozen=%s)" % (
            self.__class__.__name__,
            self.state,
            self.tid,
            self.frozen,
        )

@cython.final
cdef class MultipleValues(CachedValue):
    cdef MVCacheEntry_p entry

    @staticmethod
    cdef MultipleValues from_entry(MVCacheEntry_p entry):
        cdef MultipleValues mv = MultipleValues.__new__(MultipleValues)
        mv.entry = entry.Py_use[MVCacheEntry]()
        return mv

    def __dealloc__(self):
        release_entry(&self.entry)

    def sizeof(self):
        # At this writing, reports 72.
        return sizeof(MVCacheEntry)

    @property
    def value(self):
        return list(self)

    @property
    def key(self):
        return self.entry.key

    @property
    def frequency(self):
        return self.entry.frequency

    @property
    def weight(self):
        return self.entry.weight()

    @property
    def max_tid(self):
        return self.entry.newest_tid()

    @property
    def newest_value(self):
        return python_from_entry_p(self.entry.copy_newest_entry());


cdef class PyGeneration:
    cdef Generation* generation
    cdef readonly object __name__
    # This reference is to keep the cache object alive while we're
    # alive since it owns the Generation*.
    cdef PyCache _cache

    @staticmethod
    cdef from_generation(Generation& gen, name, cache):
        cdef PyGeneration pygen = PyGeneration.__new__(PyGeneration)
        pygen.generation = &gen
        pygen.__name__ = name
        pygen._cache = cache
        return pygen

    @property
    def generation_number(self):
        return self.generation.generation

    @property
    def limit(self):
        return self.generation.max_weight()

    @property
    def weight(self):
        return self.generation.sum_weights()

    def __len__(self):
        return self.generation.size()

    # Cython still uses __nonzero__
    def __nonzero__(self):
        return not self.generation.empty()

    def __iter__(self):
        "Not thread safe."
        if self.generation.empty():
            return ()
        it = self.generation.begin()
        while it != self.generation.end():
            yield python_from_entry(deref(it))
            preincr(it)

@cython.final
cdef class PyCache:
    cdef Cache cache
    cdef readonly size_t sets
    cdef readonly size_t hits
    cdef readonly size_t misses

    def __cinit__(self, eden, protected, probation):
        self.cache.resize(eden, protected, probation)
        self.sets = self.hits = self.misses = 0

    cpdef reset_stats(self):
        self.hits = self.sets = self.misses = 0

    @property
    def limit(self):
        return self.cache.max_weight()

    # Access to generations
    @property
    def eden(self):
        return PyGeneration.from_generation(self.cache.ring_eden,
                                            'eden',
                                            self)

    @property
    def protected(self):
        return PyGeneration.from_generation(self.cache.ring_protected,
                                            'protected',
                                            self)

    @property
    def probation(self):
        return PyGeneration.from_generation(self.cache.ring_probation,
                                            'probation',
                                            self)
    @property
    def generations(self):
        return [NoSuchGeneration(0),
                self.eden,
                self.protected,
                self.probation,]

    # Mapping operations

    def __nonzero__(self):
        # Cython still uses __nonzero__
        return self.cache.size() > 0

    def __contains__(self, OID_t key):
        return self.cache.contains(key)

    def __len__(self):
        return self.cache.size()

    cpdef CachedValue get(self, OID_t key):
        entry = self.cache.get(key)
        if not entry:
            return None
        return python_from_entry_p(entry)

    cpdef CachedValue peek(self, OID_t key):
        return self.get(key)

    cpdef object peek_item_with_tid(self, OID_t key, TID_t tid):
        value = self.cache.peek(key, tid)
        if value:
            return python_from_entry_p(value)

    cpdef bint contains_oid_with_tid(self, OID_t key, tid):
        """
        For use during cache validation. Can only be used when
        the cache only contains SingleValue entries; failure
        to do this will lead to memory leaks.
        """
        cdef TID_t native_tid = -1 if tid is None else tid
        cdef SVCacheEntry* entry = self.cache.peek(key, native_tid)
        if not entry:
            return False
        # We're guaranteed that entry is an existing object, not one we need
        # to manage the lifetime of, because of the requirement that only
        # single values can be in the cache at this time.
        return True

    def __getitem__(self, OID_t key):
        return self.get(key)

    cpdef get_item_with_tid(self, OID_t key, tid):
        cdef TID_t native_tid = -1 if tid is None else tid
        cdef SVCacheEntry* cvalue = self.cache.get(key, native_tid)

        if cvalue:
            self.hits += 1
            return SingleValue.from_entry(cvalue)

        self.misses += 1

    def __setitem__(self, OID_t key, tuple value):
        self._do_set(key, value[0], value[1])

    cdef _do_set(self, OID_t key, object state, TID_t tid):
        # Do all this down here so we don't give up the GIL.
        cdef object b_state = state if state is not None else b''
        cdef ProposedCacheEntry proposed = ProposedCacheEntry(key, tid, b_state)
        if not self.cache.contains(key): # the long way to avoid type conversion
            self.cache.add_to_eden(proposed)
        else:
            # We need to merge values.
            try:
                self.cache.store_and_make_MRU(proposed)
            except RuntimeError as e:
                raise CacheConsistencyError(str(e))

        self.sets += 1

    def __delitem__(self, OID_t key):
        self.cache.delitem(key)

    def __iter__(self):
        """
        Iterate across all the contained OID/TID pairs (the keys).

        Note that the number of items in the iterator may exceed
        the number of items in the cache due to aliasing.

        This is not thread safe.
        """
        it = self.cache.begin()
        end = self.cache.end()

        while it != end:
            oid = deref(it).key
            # This should be a small list. It mustn't persist
            # across the yield boundary, we don't own it but Cython
            # tries to keep it as a reference
            tids = list(deref(it).all_tids())
            for t in tids:
                yield (oid, t)
            preincr(it)

    def iteritems(self):
        """
        Iterate across the oid/cache_value pairs.

        Not thread safe.
        """
        it = self.cache.begin()
        end = self.cache.end()

        while it != end:
            yield (deref(it).key, python_from_entry(deref(it)))
            preincr(it)

    def keys(self):
        """
        Iterate across the OIDs in the cache.

        Not thread safe.
        """
        it = self.cache.begin()
        end = self.cache.end()

        while it != end:
            yield deref(it).key
            preincr(it)

    def values(self):
        """
        Iterate across the values in the cache.

        Not thread safe.
        """
        it = self.cache.begin()
        end = self.cache.end()

        while it != end:
            yield python_from_entry(deref(it))
            preincr(it)

    # Cache specific operations

    cpdef set_all_for_tid(self, TID_t tid_int, state_oid_iter, compress, Py_ssize_t value_limit):
        # Do all this down here so we don't give up the GIL, assuming the iter is actually a list
        # or otherwise not implemented in pure-python
        cdef OID_t oid_int
        for state_bytes, oid_int, _ in state_oid_iter:
            state_bytes = compress(state_bytes) if compress is not None else state_bytes
            state_bytes = state_bytes if state_bytes is not None else b''
            if len(state_bytes) >= value_limit:
                # This value is too big, so don't cache it.
                continue
            self._do_set(oid_int, state_bytes, tid_int)


    def add_MRUs(self, ordered_keys, return_count_only=False):
        cdef OID_t key
        cdef TID_t tid
        cdef int i
        cdef int number_nodes = len(ordered_keys)
        cdef TempCacheFiller filler
        if not number_nodes:
            return 0 if return_count_only else ()

        filler = TempCacheFiller(number_nodes)

        for i, (key, (state, tid, frozen, frequency)) in enumerate(ordered_keys):
            state = state or b''
            filler.set_once_at(i, key, tid, state, frozen, frequency)


        # We're done with ordered_keys, free its memory
        ordered_keys = None

        added_oids = self.cache.add_many(filler)

        # Things that didn't get added have -1 for their generation.
        if return_count_only:
            return added_oids.size()

        result = [self.get(key) for key in added_oids]

        return result

    def age_frequencies(self):
        self.cache.age_frequencies()

    def delitems(self, oids_tids):
        """
        For each OID/TID pair in the items, remove all cached values
        for OID that are older than TID.
        """
        cdef OID_t oid
        cdef TID_t tid

        for oid, tid in iteroiditems(oids_tids):
            self.cache.delitem(oid, tid)

    def del_oids(self, oids):
        """
        For each oid in OIDs, remove it.
        """
        for oid in oids:
            self.cache.delitem(oid)

    def freeze(self, oids_tids):
        # The idea is to *move* the data, or make it available,
        # *without* copying it.
        cdef OID_t oid
        cdef TID_t tid

        for oid, tid in iteroiditems(oids_tids):
            self.cache.freeze(oid, tid)

    @property
    def weight(self):
        return self.cache.weight()

# Local Variables:
# flycheck-cython-cplus: t
# End:
